#
# Copyright (C) 2007,2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""Common functions for various parts of Deluge to use."""

import base64
import binascii
import functools
import glob
import logging
import numbers
import os
import platform
import re
import socket
import subprocess
import sys
import tarfile
import time
from contextlib import closing
from datetime import datetime
from importlib import resources
from io import BytesIO
from pathlib import Path
from urllib.parse import unquote_plus, urljoin
from urllib.request import pathname2url

from deluge.decorators import deprecated
from deluge.error import InvalidPathError

try:
    from importlib.metadata import distribution
except ImportError:
    from pkg_resources import get_distribution as distribution


try:
    import chardet
except ImportError:
    chardet = None

# Windows workaround for HTTPS requests requiring certificate authority bundle.
# see: https://twistedmatrix.com/trac/ticket/9209
if platform.system() in ('Windows', 'Microsoft'):
    from certifi import where

    os.environ['SSL_CERT_FILE'] = where()

try:
    import ifaddr
except ImportError:
    ifaddr = None


if platform.system() not in ('Windows', 'Microsoft', 'Darwin'):
    # gi makes dbus available on Window but don't import it as unused.
    try:
        import dbus
    except ImportError:
        dbus = None
    try:
        import distro
    except ImportError:
        distro = None

log = logging.getLogger(__name__)

TORRENT_STATE = [
    'Allocating',
    'Checking',
    'Downloading',
    'Seeding',
    'Paused',
    'Error',
    'Queued',
    'Moving',
]

# The output formatting for json.dump
JSON_FORMAT = {'indent': 4, 'sort_keys': True, 'ensure_ascii': False}

DBUS_FM_ID = 'org.freedesktop.FileManager1'
DBUS_FM_PATH = '/org/freedesktop/FileManager1'

# Retained for plugin backward compatibility
PY2 = False


def get_version():
    """The program version from the egg metadata.

    Returns:
        str: The version of Deluge.
    """
    return distribution('Deluge').version


def get_default_config_dir(filename=None):
    """
    :param filename: if None, only the config path is returned, if provided,
                     a path including the filename will be returned
    :type filename: string
    :returns: a file path to the config directory and optional filename
    :rtype: string

    """

    if windows_check():

        def save_config_path(resource):
            app_data_path = os.environ.get('APPDATA')
            if not app_data_path:
                import winreg

                hkey = winreg.OpenKey(
                    winreg.HKEY_CURRENT_USER,
                    'Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders',
                )
                app_data_reg = winreg.QueryValueEx(hkey, 'AppData')
                app_data_path = app_data_reg[0]
                winreg.CloseKey(hkey)
            return os.path.join(app_data_path, resource)

    else:
        from xdg.BaseDirectory import save_config_path
    if not filename:
        filename = ''
    try:
        return decode_bytes(os.path.join(save_config_path('deluge'), filename))
    except OSError as ex:
        log.error('Unable to use default config directory, exiting... (%s)', ex)
        sys.exit(1)


def get_default_download_dir():
    """
    :returns: the default download directory
    :rtype: string

    """
    download_dir = ''
    if not windows_check():
        from xdg.BaseDirectory import xdg_config_home

        try:
            user_dirs_path = os.path.join(xdg_config_home, 'user-dirs.dirs')
            with open(user_dirs_path, encoding='utf8') as _file:
                for line in _file:
                    if not line.startswith('#') and line.startswith('XDG_DOWNLOAD_DIR'):
                        download_dir = os.path.expandvars(
                            line.partition('=')[2].rstrip().strip('"')
                        )
                        break
        except OSError:
            pass

    if not download_dir:
        download_dir = os.path.join(os.path.expanduser('~'), 'Downloads')
    return download_dir


def archive_files(arc_name, filepaths, message=None, rotate=10):
    """Compress a list of filepaths into timestamped tarball in config dir.

    The archiving config directory is 'archive'.

    Args:
        arc_name (str): The archive output filename (appended with timestamp).
        filepaths (list): A list of the files to be archived into tarball.

    Returns:
        str: The full archive filepath.

    """

    from deluge.configmanager import get_config_dir

    # Set archive compression to lzma
    arc_comp = 'xz'

    archive_dir = os.path.join(get_config_dir(), 'archive')
    timestamp = datetime.now().replace(microsecond=0).isoformat().replace(':', '-')
    arc_filepath = os.path.join(
        archive_dir, arc_name + '-' + timestamp + '.tar.' + arc_comp
    )

    if not os.path.exists(archive_dir):
        os.makedirs(archive_dir)
    else:
        all_arcs = glob.glob(os.path.join(archive_dir, arc_name) + '*')
        if len(all_arcs) >= rotate:
            log.warning(
                'Too many existing archives for %s. Deleting oldest archive.', arc_name
            )
            os.remove(sorted(all_arcs)[0])

    try:
        with tarfile.open(arc_filepath, 'w:' + arc_comp) as tar:
            for filepath in filepaths:
                if not os.path.isfile(filepath):
                    continue
                tar.add(filepath, arcname=os.path.basename(filepath))
            if message:
                with closing(BytesIO(message.encode('utf8'))) as fobj:
                    tarinfo = tarfile.TarInfo('archive_message.txt')
                    tarinfo.size = len(fobj.getvalue())
                    tarinfo.mtime = time.time()
                    tar.addfile(tarinfo, fileobj=fobj)
    except OSError:
        log.error('Problem occurred archiving filepaths: %s', filepaths)
        return False
    else:
        return arc_filepath


def windows_check():
    """
    Checks if the current platform is Windows

    :returns: True or False
    :rtype: bool

    """
    return platform.system() in ('Windows', 'Microsoft')


def vista_check():
    """
    Checks if the current platform is Windows Vista

    :returns: True or False
    :rtype: bool

    """
    return platform.release() == 'Vista'


def osx_check():
    """
    Checks if the current platform is Mac OS X

    :returns: True or False
    :rtype: bool

    """
    return platform.system() == 'Darwin'


def linux_check():
    """
    Checks if the current platform is Linux

    :returns: True or False
    :rtype: bool

    """
    return platform.system() == 'Linux'


def get_os_version():
    """Parse and return the os version information.

    Converts the platform ver tuple to a string.

    Returns:
        str: The os version info.

    """
    if windows_check():
        os_version = platform.win32_ver()
    elif osx_check():
        os_version = list(platform.mac_ver())
        os_version[1] = ''  # versioninfo always empty.
    elif distro:
        os_version = (distro.name(), distro.version(), distro.codename())
    else:
        os_version = (platform.release(),)

    return ' '.join(filter(None, os_version))


def get_pixmap(fname):
    """
    Provides easy access to files in the deluge/ui/data/pixmaps folder within the Deluge egg

    :param fname: the filename to look for
    :type fname: string
    :returns: a path to a pixmap file included with Deluge
    :rtype: string

    """
    return resource_filename('deluge', os.path.join('ui', 'data', 'pixmaps', fname))


def resource_filename(module: str, path: str) -> str:
    """Get filesystem path for a non-python resource.

    Abstracts getting module resource files. Originally created to
    workaround pkg_resources.resource_filename limitations with
    multiple Deluge packages installed.
    """
    path = Path(path)

    try:
        with resources.as_file(resources.files(module) / path) as resource_file:
            return str(resource_file)
    except AttributeError:
        # Python <= 3.8
        with resources.path(module, path.parts[0]) as resource_file:
            return str(resource_file.joinpath(*path.parts[1:]))


def open_file(path, timestamp=None):
    """Opens a file or folder using the system configured program.

    Args:
        path (str): The path to the file or folder to open.
        timestamp (int, optional): An event request timestamp.

    """
    if windows_check():
        os.startfile(path)
    elif osx_check():
        subprocess.Popen(['open', path])
    else:
        if timestamp is None:
            timestamp = int(time.time())
        env = os.environ.copy()
        env['DESKTOP_STARTUP_ID'] = '%s-%u-%s-xdg_open_TIME%d' % (
            os.path.basename(sys.argv[0]),
            os.getpid(),
            os.uname()[1],
            timestamp,
        )
        subprocess.Popen(['xdg-open', '%s' % path], env=env)


def show_file(path, timestamp=None):
    """Shows (highlights) a file or folder using the system configured file manager.

    Args:
        path (str): The path to the file or folder to show.
        timestamp (int, optional): An event request timestamp.

    """
    if windows_check():
        subprocess.Popen(['explorer', '/select,', path])
    elif osx_check():
        subprocess.Popen(['open', '-R', path])
    else:
        if timestamp is None:
            timestamp = int(time.time())
        startup_id = '%s_%u_%s-dbus_TIME%d TIMESTAMP=%d' % (
            os.path.basename(sys.argv[0]),
            os.getpid(),
            os.uname()[1],
            timestamp,
            timestamp,
        )

        if dbus:
            bus = dbus.SessionBus()
            try:
                filemanager1 = bus.get_object(DBUS_FM_ID, DBUS_FM_PATH)
            except dbus.exceptions.DBusException as ex:
                log.debug('Unable to get dbus file manager: %s', ex)
                # Fallback to xdg-open
            else:
                paths = [urljoin('file:', pathname2url(path))]
                filemanager1.ShowItems(paths, startup_id, dbus_interface=DBUS_FM_ID)
                return

        env = os.environ.copy()
        env['DESKTOP_STARTUP_ID'] = startup_id.replace('dbus', 'xdg-open')
        # No option in xdg to highlight a file so just open parent folder.
        subprocess.Popen(['xdg-open', os.path.dirname(path.rstrip('/'))], env=env)


def open_url_in_browser(url):
    """
    Opens a URL in the desktop's default browser

    :param url: the URL to open
    :type url: string

    """
    import webbrowser

    webbrowser.open(url)


# Formatting text functions
byte_txt = 'B'
kib_txt = 'KiB'
mib_txt = 'MiB'
gib_txt = 'GiB'
tib_txt = 'TiB'
kib_txt_short = 'K'
mib_txt_short = 'M'
gib_txt_short = 'G'
tib_txt_short = 'T'


def translate_size_units():
    """For performance reasons these units are translated outside the function"""

    global byte_txt, kib_txt, mib_txt, gib_txt, tib_txt
    global kib_txt_short, mib_txt_short, gib_txt_short, tib_txt_short

    byte_txt = _('B')
    kib_txt = _('KiB')
    mib_txt = _('MiB')
    gib_txt = _('GiB')
    tib_txt = _('TiB')
    kib_txt_short = _('K')
    mib_txt_short = _('M')
    gib_txt_short = _('G')
    tib_txt_short = _('T')


def fsize(fsize_b, precision=1, shortform=False):
    """Formats the bytes value into a string with KiB, MiB, GiB or TiB units.

    Args:
        fsize_b (int): The filesize in bytes.
        precision (int): The output float precision, 1 by default.
        shortform (bool): The output short|long form, False (long form) by default.

    Returns:
        str: A formatted string in KiB, MiB, GiB or TiB units.

    Examples:
        >>> fsize(112245)
        '109.6 KiB'
        >>> fsize(112245, precision=0)
        '110 KiB'
        >>> fsize(112245, shortform=True)
        '109.6 K'

    Note:
        This function has been refactored for performance with the
        fsize units being translated outside the function.

        Notice that short forms K|M|G|T are synonymous here with
        KiB|MiB|GiB|TiB. They are powers of 1024, not 1000.

    """

    if fsize_b >= 1024**4:
        return '%.*f %s' % (
            precision,
            fsize_b / 1024**4,
            tib_txt_short if shortform else tib_txt,
        )
    elif fsize_b >= 1024**3:
        return '%.*f %s' % (
            precision,
            fsize_b / 1024**3,
            gib_txt_short if shortform else gib_txt,
        )
    elif fsize_b >= 1024**2:
        return '%.*f %s' % (
            precision,
            fsize_b / 1024**2,
            mib_txt_short if shortform else mib_txt,
        )
    elif fsize_b >= 1024:
        return '%.*f %s' % (
            precision,
            fsize_b / 1024,
            kib_txt_short if shortform else kib_txt,
        )
    else:
        return '%d %s' % (fsize_b, byte_txt)


def fpcnt(dec, precision=2):
    """Formats a string to display a percentage with <precision> places.

    Args:
        dec (float): The ratio in the range [0.0, 1.0].
        precision (int): The output float precision, 2 by default.

    Returns:
        str: A formatted string representing a percentage.

    Examples:
        >>> fpcnt(0.9311)
        '93.11%'
        >>> fpcnt(0.9311, precision=0)
        '93%'

    """

    pcnt = dec * 100
    if pcnt == 0 or pcnt == 100:
        precision = 0
    return '%.*f%%' % (precision, pcnt)


def fspeed(bps, precision=1, shortform=False):
    """Formats a string to display a transfer speed.

    Args:
        bps (int): The speed in bytes per second.
        precision (int): The output float precision, 1 by default.
        shortform (bool): The output short|long form, False (long form) by default.

    Returns:
        str: A formatted string representing transfer speed.

    Examples:
        >>> fspeed(43134)
        '42.1 KiB/s'

    Note:
        Notice that short forms K|M|G|T are synonymous here with
        KiB|MiB|GiB|TiB. They are powers of 1024, not 1000.

    """

    if bps < 1024**2:
        return '%.*f %s' % (
            precision,
            bps / 1024,
            _('K/s') if shortform else _('KiB/s'),
        )
    elif bps < 1024**3:
        return '%.*f %s' % (
            precision,
            bps / 1024**2,
            _('M/s') if shortform else _('MiB/s'),
        )
    elif bps < 1024**4:
        return '%.*f %s' % (
            precision,
            bps / 1024**3,
            _('G/s') if shortform else _('GiB/s'),
        )
    else:
        return '%.*f %s' % (
            precision,
            bps / 1024**4,
            _('T/s') if shortform else _('TiB/s'),
        )


def fpeer(num_peers, total_peers):
    """Formats a string to show 'num_peers' ('total_peers').

    Args:
        num_peers (int): The number of connected peers.
        total_peers (int): The total number of peers.

    Returns:
        str: A formatted string 'num_peers (total_peers)' or if total_peers < 0, just 'num_peers'.

    Examples:
        >>> fpeer(10, 20)
        '10 (20)'
        >>> fpeer(10, -1)
        '10'

    """
    if total_peers > -1:
        return f'{num_peers:d} ({total_peers:d})'
    else:
        return f'{num_peers:d}'


def ftime(secs):
    """Formats a string to show time in a human readable form.

    Args:
        secs (int or float): The number of seconds.

    Returns:
        str: A formatted time string or empty string if value is 0.

    Examples:
        >>> ftime(23011)
        '6h 23m'

    Note:
        This function has been refactored for performance.

    """

    # Handle floats by truncating to an int
    secs = int(secs)
    if secs <= 0:
        time_str = ''
    elif secs < 60:
        time_str = f'{secs}s'
    elif secs < 3600:
        time_str = f'{secs // 60}m {secs % 60}s'
    elif secs < 86400:
        time_str = f'{secs // 3600}h {secs // 60 % 60}m'
    elif secs < 604800:
        time_str = f'{secs // 86400}d {secs // 3600 % 24}h'
    elif secs < 31449600:
        time_str = f'{secs // 604800}w {secs // 86400 % 7}d'
    else:
        time_str = f'{secs // 31449600}y {secs // 604800 % 52}w'
    return time_str


def fdate(seconds, date_only=False, precision_secs=False):
    """Formats a date time string in the locale's date representation based on the system's timezone.

    Args:
        seconds (float): Time in seconds since the Epoch.
        date_only (bool): Whether to include only the date, False by default.
        precision_secs (bool): Include seconds in time format, False by default.

    Returns:
        str: A string in the locale's datetime representation or "" if seconds < 0

    """

    if seconds < 0:
        return ''
    time_format = '%x %X' if precision_secs else '%x %H:%M'
    if date_only:
        time_format = time_format.split()[0]
    return time.strftime(time_format, time.localtime(seconds))


def tokenize(text):
    """
    Tokenize a text into numbers and strings.

    Args:
        text (str): The text to tokenize (a string).

    Returns:
        list: A list of strings and/or numbers.

    Note:
        This function is used to implement robust tokenization of user input
        It automatically coerces integer and floating point numbers, ignores
        whitespace and knows how to separate numbers from strings even without
        whitespace.

        Possible optimization: move the 2 regexes outside of function.

    """
    tokenized_input = []
    for token in re.split(r'(\d+(?:\.\d+)?)', text):
        token = token.strip()
        if re.match(r'\d+\.\d+', token):
            tokenized_input.append(float(token))
        elif token.isdigit():
            tokenized_input.append(int(token))
        elif token:
            tokenized_input.append(token)
    return tokenized_input


size_units = [
    {'prefix': 'b', 'divider': 1, 'singular': 'byte', 'plural': 'bytes'},
    {'prefix': 'KiB', 'divider': 1024**1},
    {'prefix': 'MiB', 'divider': 1024**2},
    {'prefix': 'GiB', 'divider': 1024**3},
    {'prefix': 'TiB', 'divider': 1024**4},
    {'prefix': 'PiB', 'divider': 1024**5},
    {'prefix': 'k', 'divider': 1000**1},
    {'prefix': 'm', 'divider': 1000**2},
    {'prefix': 'g', 'divider': 1000**3},
    {'prefix': 't', 'divider': 1000**4},
    {'prefix': 'p', 'divider': 1000**5},
    {'prefix': 'KB', 'divider': 1000**1},
    {'prefix': 'MB', 'divider': 1000**2},
    {'prefix': 'GB', 'divider': 1000**3},
    {'prefix': 'TB', 'divider': 1000**4},
    {'prefix': 'PB', 'divider': 1000**5},
]


class InvalidSize(Exception):
    pass


def parse_human_size(size):
    """
    Parse a human readable data size and return the number of bytes.

    Args:
        size (str): The human readable file size to parse (a string).

    Returns:
        int: The corresponding size in bytes.

    Raises:
        InvalidSize: when the input can't be parsed.

    """
    tokens = tokenize(size)
    if tokens and isinstance(tokens[0], numbers.Number):
        # If the input contains only a number, it's assumed to be the number of bytes.
        if len(tokens) == 1:
            return int(tokens[0])
        # Otherwise we expect to find two tokens: A number and a unit.
        if len(tokens) == 2:
            try:
                normalized_unit = tokens[1].lower()
            except AttributeError:
                pass
            else:
                # Try to match the first letter of the unit.
                for unit in size_units:
                    if normalized_unit.startswith(unit['prefix'].lower()):
                        return int(tokens[0] * unit['divider'])
    # We failed to parse the size specification.
    msg = 'Failed to parse size! (input %r was tokenized as %r)'
    raise InvalidSize(msg % (size, tokens))


def anchorify_urls(text: str) -> str:
    """
    Wrap all occurrences of text URLs with HTML
    """
    url_pattern = r'((htt)|(ft)|(ud))ps?://\S+'
    html_href_pattern = r'<a href="\g<0>">\g<0></a>'

    return re.sub(url_pattern, html_href_pattern, text)


def is_url(url):
    """
    A simple test to check if the URL is valid

    :param url: the URL to test
    :type url: string
    :returns: True or False
    :rtype: bool

    :Example:

    >>> is_url('http://deluge-torrent.org')
    True

    """
    if not url:
        return False

    return url.partition('://')[0] in ('http', 'https', 'ftp', 'udp')


def is_infohash(infohash):
    """
    A check to determine if a string is a valid infohash.

    Args:
        infohash (str): The string to check.

    Returns:
        bool: True if valid infohash, False otherwise.

    """
    if not infohash:
        return False

    return len(infohash) == 40 and infohash.isalnum()


MAGNET_SCHEME = 'magnet:?'
XT_BTIH_PARAM = 'xt=urn:btih:'
DN_PARAM = 'dn='
TR_PARAM = 'tr='
TR_TIER_PARAM = 'tr.'
TR_TIER_REGEX = re.compile(r'^tr.(\d+)=(\S+)')


def is_magnet(uri):
    """
    A check to determine if a URI is a valid bittorrent magnet URI

    :param uri: the URI to check
    :type uri: string
    :returns: True or False
    :rtype: bool

    :Example:

    >>> is_magnet('magnet:?xt=urn:btih:SU5225URMTUEQLDXQWRB2EQWN6KLTYKN')
    True

    """
    if not uri:
        return False

    return uri.startswith(MAGNET_SCHEME) and XT_BTIH_PARAM in uri


def get_magnet_info(uri):
    """Parse torrent information from magnet link.

    Args:
        uri (str): The magnet link.

    Returns:
        dict: Information about the magnet link.

        Format of the magnet dict::

            {
                "name": the torrent name,
                "info_hash": the torrents info_hash,
                "files_tree": empty value for magnet links
            }

    """

    if not uri.startswith(MAGNET_SCHEME):
        return {}

    name = None
    info_hash = None
    trackers = {}
    tier = 0
    for param in uri[len(MAGNET_SCHEME) :].split('&'):
        if param.startswith(XT_BTIH_PARAM):
            xt_hash = param[len(XT_BTIH_PARAM) :]
            if len(xt_hash) == 32:
                try:
                    infohash_str = base64.b32decode(xt_hash.upper())
                except TypeError as ex:
                    log.debug('Invalid base32 magnet hash: %s, %s', xt_hash, ex)
                    break
                info_hash = binascii.hexlify(infohash_str).decode()
            elif is_infohash(xt_hash):
                info_hash = xt_hash.lower()
            else:
                break
        elif param.startswith(DN_PARAM):
            name = unquote_plus(param[len(DN_PARAM) :])
        elif param.startswith(TR_PARAM):
            tracker = unquote_plus(param[len(TR_PARAM) :])
            trackers[tracker] = tier
            tier += 1
        elif param.startswith(TR_TIER_PARAM):
            tracker_match = re.match(TR_TIER_REGEX, param)
            if not tracker_match:
                continue

            tier, tracker = tracker_match.groups()
            tracker = unquote_plus(tracker)
            trackers[tracker] = int(tier)

    if info_hash:
        if not name:
            name = info_hash
        return {
            'name': name,
            'info_hash': info_hash,
            'files_tree': '',
            'trackers': trackers,
        }
    else:
        return {}


def create_magnet_uri(infohash, name=None, trackers=None):
    """Creates a magnet URI

    Args:
        infohash (str): The info-hash of the torrent.
        name (str, optional): The name of the torrent.
        trackers (list or dict, optional): A list of trackers or a dict or some {tracker: tier} pairs.

    Returns:
        str: A magnet URI string.

    """
    try:
        infohash = binascii.unhexlify(infohash)
    except TypeError:
        infohash.encode('utf-8')

    uri = [MAGNET_SCHEME, XT_BTIH_PARAM, base64.b32encode(infohash).decode('utf-8')]
    if name:
        uri.extend(['&', DN_PARAM, name])
    if trackers:
        try:
            for tracker in sorted(trackers, key=trackers.__getitem__):
                uri.extend(['&', 'tr.%d=' % trackers[tracker], tracker])
        except TypeError:
            for tracker in trackers:
                uri.extend(['&', TR_PARAM, tracker])

    return ''.join(uri)


def get_path_size(path):
    """
    Gets the size in bytes of 'path'

    :param path: the path to check for size
    :type path: string
    :returns: the size in bytes of the path or -1 if the path does not exist
    :rtype: int

    """
    if not os.path.exists(path):
        return -1

    if os.path.isfile(path):
        return os.path.getsize(path)

    dir_size = 0
    for p, dummy_dirs, files in os.walk(path):
        for _file in files:
            filename = os.path.join(p, _file)
            dir_size += os.path.getsize(filename)
    return dir_size


def free_space(path):
    """
    Gets the free space available at 'path'

    :param path: the path to check
    :type path: string
    :returns: the free space at path in bytes
    :rtype: int

    :raises InvalidPathError: if the path is not valid

    """
    if not path or not os.path.exists(path):
        raise InvalidPathError('%s is not a valid path' % path)

    if windows_check():
        from win32file import GetDiskFreeSpaceEx

        return GetDiskFreeSpaceEx(path)[0]
    else:
        disk_data = os.statvfs(path.encode('utf8'))
        block_size = disk_data.f_frsize
        return disk_data.f_bavail * block_size


def is_interface(interface):
    """Check if interface is a valid IP or network adapter.

    Args:
        interface (str): The IP or interface name to test.

    Returns:
        bool: Whether interface is valid is not.

    Examples:
        Windows:
        >>> is_interface('{7A30AE62-23ZA-3744-Z844-A5B042524871}')
        >>> is_interface('127.0.0.1')
        True
        Linux:
        >>> is_interface('lo')
        >>> is_interface('127.0.0.1')
        True

    """
    return is_ip(interface) or is_interface_name(interface)


def is_ip(ip):
    """A test to see if 'ip' is a valid IPv4 or IPv6 address.

    Args:
        ip (str): The IP to test.

    Returns:
        bool: Whether IP is valid is not.

    Examples:
        >>> is_ip("192.0.2.0")
        True
        >>> is_ip("2001:db8::")
        True

    """

    return is_ipv4(ip) or is_ipv6(ip)


def is_ipv4(ip):
    """A test to see if 'ip' is a valid IPv4 address.

    Args:
        ip (str): The IP to test.

    Returns:
        bool: Whether IP is valid is not.

    Examples:
        >>> is_ipv4("192.0.2.0")
        True

    """

    try:
        socket.inet_pton(socket.AF_INET, ip)
    except OSError:
        return False
    else:
        return True


def is_ipv6(ip):
    """A test to see if 'ip' is a valid IPv6 address.

    Args:
        ip (str): The IP to test.

    Returns:
        bool: Whether IP is valid is not.

    Examples:
        >>> is_ipv6("2001:db8::")
        True

    """

    try:
        socket.inet_pton(socket.AF_INET6, ip)
    except OSError:
        return False
    else:
        return True


def is_interface_name(name):
    """Returns True if an interface name exists.

    Args:
        name (str): The Interface to test. eg. eth0 linux. GUID on Windows.

    Returns:
        bool: Whether name is valid or not.

    Examples:
        >>> is_interface_name("eth0")
        True
        >>> is_interface_name("{7A30AE62-23ZA-3744-Z844-A5B042524871}")
        True

    """

    if not windows_check():
        try:
            socket.if_nametoindex(name)
        except OSError:
            pass
        else:
            return True

    if ifaddr:
        try:
            adapters = ifaddr.get_adapters()
        except OSError:
            return True
        else:
            return any([name == a.name for a in adapters])

    if windows_check():
        regex = '^{[0-9A-Z]{8}-([0-9A-Z]{4}-){3}[0-9A-Z]{12}}$'
        return bool(re.search(regex, str(name)))

    return True


def decode_bytes(byte_str, encoding='utf8'):
    """Decodes a byte string and return unicode.

    If it cannot decode using `encoding` then it will try latin1,
    and if that fails, try to detect the string encoding. If that fails,
    decode with ignore.

    Args:
        byte_str (bytes): The byte string to decode.
        encoding (str): The encoding to try first when decoding.

    Returns:
        str: A unicode string.

    """
    if not byte_str:
        return ''
    elif not isinstance(byte_str, bytes):
        return byte_str

    encodings = [lambda: ('utf8', 'strict'), lambda: ('iso-8859-1', 'strict')]
    if chardet:
        encodings.append(lambda: (chardet.detect(byte_str)['encoding'], 'strict'))
    encodings.append(lambda: (encoding, 'ignore'))

    if encoding.lower() not in ['utf8', 'utf-8']:
        encodings.insert(0, lambda: (encoding, 'strict'))

    for enc in encodings:
        try:
            return byte_str.decode(*enc())
        except UnicodeDecodeError:
            pass
    return ''


@deprecated
def decode_string(byte_str, encoding='utf8'):
    """Deprecated: Use decode_bytes"""
    return decode_bytes(byte_str, encoding)


@deprecated
def utf8_encoded(str_, encoding='utf8'):
    """Deprecated: Use encode or decode_bytes if needed"""
    return decode_bytes(str_, encoding).encode('utf8')


def utf8_encode_structure(data):
    """Recursively convert all unicode keys and values in a data structure to utf8.

    e.g. converting keys and values for a dict with nested dicts and lists etc.

    Args:
        data (any): This can be any structure, dict, list or tuple.

    Returns:
        input type: The data with unicode keys and values converted to utf8.

    """
    if isinstance(data, (list, tuple)):
        return type(data)([utf8_encode_structure(d) for d in data])
    elif isinstance(data, dict):
        return {
            utf8_encode_structure(k): utf8_encode_structure(v) for k, v in data.items()
        }
    elif not isinstance(data, bytes):
        try:
            return data.encode('utf8')
        except AttributeError:
            pass
    return data


@functools.total_ordering
class VersionSplit:
    """
    Used for comparing version numbers.

    :param ver: the version
    :type ver: string

    """

    def __init__(self, ver):
        version_re = re.compile(
            r"""
        ^
        (?P<version>\d+\.\d+)          # minimum 'N.N'
        (?P<extraversion>(?:\.\d+)*)   # any number of extra '.N' segments
        (?:
            (?P<prerel>[abc]|rc)       # 'a'=alpha, 'b'=beta, 'c'=release candidate
                                       # 'rc'= alias for release candidate
            (?P<prerelversion>\d+(?:\.\d+)*)
        )?
        (?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?
        $""",
            re.VERBOSE,
        )

        # Check for PEP 386 compliant version
        match = re.search(version_re, ver)
        if match:
            group = [(x if x is not None else '') for x in match.group(1, 2, 3, 4, 8)]
            vs = [''.join(group[0:2]), ''.join(group[2:4]), group[4].lstrip('.')]
        else:
            ver = ver.lower()
            vs = ver.replace('_', '-').split('-')

        self.version = [int(x) for x in vs[0].split('.') if x.isdigit()]
        self.version_string = ''.join(str(x) for x in vs[0].split('.') if x.isdigit())
        self.suffix = None
        self.dev = None
        if len(vs) > 1:
            if vs[1].startswith(('rc', 'a', 'b', 'c')):
                self.suffix = vs[1]
            if vs[-1].startswith('dev'):
                try:
                    # Store only the dev numeral.
                    self.dev = int(vs[-1].rsplit('dev')[1])
                except ValueError:
                    # Implicit dev numeral is 0.
                    self.dev = 0

    def get_comparable_versions(self, other):
        """
        Returns a 2-tuple of lists for use in the comparison
        methods.
        """
        # PEP 386 versions with .devN precede release version so default
        # non-dev versions to infinity while dev versions are ints.
        self.dev = float('inf') if self.dev is None else self.dev
        other.dev = float('inf') if other.dev is None else other.dev
        # If there is no suffix we use z because we want final
        # to appear after alpha, beta, and rc alphabetically.
        v1 = [self.version, self.suffix or 'z', self.dev]
        v2 = [other.version, other.suffix or 'z', other.dev]

        return (v1, v2)

    def __eq__(self, other):
        v1, v2 = self.get_comparable_versions(other)
        return v1 == v2

    def __lt__(self, other):
        v1, v2 = self.get_comparable_versions(other)
        return v1 < v2


# Common AUTH stuff
AUTH_LEVEL_NONE = 0
AUTH_LEVEL_READONLY = 1
AUTH_LEVEL_NORMAL = 5
AUTH_LEVEL_ADMIN = 10
AUTH_LEVEL_DEFAULT = AUTH_LEVEL_NORMAL


def create_auth_file():
    import stat

    import deluge.configmanager

    auth_file = deluge.configmanager.get_config_dir('auth')
    # Check for auth file and create if necessary
    if not os.path.exists(auth_file):
        with open(auth_file, 'w', encoding='utf8') as _file:
            _file.flush()
            os.fsync(_file.fileno())
        # Change the permissions on the file so only this user can read/write it
        os.chmod(auth_file, stat.S_IREAD | stat.S_IWRITE)


def create_localclient_account(append=False):
    import random
    from hashlib import sha1 as sha

    import deluge.configmanager

    auth_file = deluge.configmanager.get_config_dir('auth')
    if not os.path.exists(auth_file):
        create_auth_file()

    with open(auth_file, 'a' if append else 'w', encoding='utf8') as _file:
        _file.write(
            ':'.join(
                [
                    'localclient',
                    sha(str(random.random()).encode('utf8')).hexdigest(),
                    str(AUTH_LEVEL_ADMIN),
                ]
            )
            + '\n'
        )
        _file.flush()
        os.fsync(_file.fileno())


def get_localhost_auth():
    """Grabs the localclient auth line from the 'auth' file and creates a localhost URI.

    Returns:
        tuple: With the username and password to login as.
    """
    from deluge.configmanager import get_config_dir

    auth_file = get_config_dir('auth')
    if not os.path.exists(auth_file):
        from deluge.common import create_localclient_account

        create_localclient_account()

    with open(auth_file, encoding='utf8') as auth:
        for line in auth:
            line = line.strip()
            if line.startswith('#') or not line:
                # This is a comment or blank line
                continue

            lsplit = line.split(':')

            if len(lsplit) == 2:
                username, password = lsplit
            elif len(lsplit) == 3:
                username, password, level = lsplit
            else:
                log.error('Your auth file is malformed: Incorrect number of fields!')
                continue

            if username == 'localclient':
                return (username, password)


def set_env_variable(name, value):
    """
    :param name: environment variable name
    :param value: environment variable value

    This function ensures that changes to an environment variable are applied
    to each copy of the environment variables used by a process. Starting from
    Python 2.4, os.environ changes only apply to the copy Python keeps (os.environ)
    and are no longer automatically applied to the other copies for the process.

    On Microsoft Windows, each process has multiple copies of the environment
    variables, one managed by the OS and one managed by the C library. We also
    need to take care of the fact that the C library used by Python is not
    necessarily the same as the C library used by pygtk and friends. This because
    the latest releases of pygtk and friends are built with mingw32 and are thus
    linked against msvcrt.dll. The official gtk+ binaries have always been built
    in this way.

    Basen on _putenv in TransUtils.py from sourceforge project gramps
    http://sourceforge.net/p/gramps/code/HEAD/tree/branches/maintenance/gramps32/src/TransUtils.py
    """
    # Update Python's copy of the environment variables
    os.environ[name] = value

    if windows_check():
        from ctypes import cdll, windll

        # Update the copy maintained by Windows (so SysInternals Process Explorer sees it)
        result = windll.kernel32.SetEnvironmentVariableW(name, value)
        if result == 0:
            log.info(
                "Failed to set Env Var '%s' (kernel32.SetEnvironmentVariableW)", name
            )
        else:
            log.debug(
                "Set Env Var '%s' to '%s' (kernel32.SetEnvironmentVariableW)",
                name,
                value,
            )

        # Update the copy maintained by msvcrt (used by gtk+ runtime)
        result = cdll.msvcrt._wputenv(f'{name}={value}')
        if result != 0:
            log.info("Failed to set Env Var '%s' (msvcrt._putenv)", name)
        else:
            log.debug("Set Env Var '%s' to '%s' (msvcrt._putenv)", name, value)


def run_profiled(func, *args, **kwargs):
    """
    Profile a function with cProfile

    Args:
        func (func): The function to profile
        *args (tuple): The arguments to pass to the function
        do_profile (bool, optional): If profiling should be performed. Defaults to True.
        output_file (str, optional): Filename to save profile results. If None, print to stdout.
                                     Defaults to None.
    """
    if kwargs.get('do_profile', True) is not False:
        import cProfile

        profiler = cProfile.Profile()

        def on_shutdown():
            output_file = kwargs.get('output_file', None)
            if output_file:
                profiler.dump_stats(output_file)
                log.info('Profile stats saved to %s', output_file)
                print('Profile stats saved to %s' % output_file)
            else:
                import pstats
                from io import StringIO

                strio = StringIO()
                ps = pstats.Stats(profiler, stream=strio).sort_stats('cumulative')
                ps.print_stats()
                print(strio.getvalue())

        try:
            return profiler.runcall(func, *args)
        finally:
            on_shutdown()
    else:
        return func(*args)


def is_process_running(pid):
    """
    Verify if the supplied pid is a running process.

    Args:
        pid (int): The pid to check.

    Returns:
        bool: True if pid is a running process, False otherwise.

    """

    if windows_check():
        from win32process import EnumProcesses

        return pid in EnumProcesses()
    else:
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge.decorators import overrides
from deluge.ui.console.modes.basemode import InputKeyHandler, move_cursor
from deluge.ui.console.modes.torrentlist.torrentactions import torrent_actions_popup
from deluge.ui.console.utils import curses_util as util

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)
QUEUE_MODE_HELP_STR = """
Change queue position of selected torrents

{!info!}'+'{!normal!} - {|indent_pos:|}Move up
{!info!}'-'{!normal!} - {|indent_pos:|}Move down

{!info!}'Home'{!normal!} - {|indent_pos:|}Move to top
{!info!}'End'{!normal!} - {|indent_pos:|}Move to bottom

"""
SEARCH_EMPTY = 0
SEARCH_FAILING = 1
SEARCH_SUCCESS = 2
SEARCH_START_REACHED = 3
SEARCH_END_REACHED = 4
SEARCH_FORMAT = {
    SEARCH_EMPTY: '{!black,white!}Search torrents: %s{!black,white!}',
    SEARCH_SUCCESS: '{!black,white!}Search torrents: {!black,green!}%s{!black,white!}',
    SEARCH_FAILING: '{!black,white!}Search torrents: {!black,red!}%s{!black,white!}',
    SEARCH_START_REACHED: '{!black,white!}Search torrents: {!black,yellow!}%s{!black,white!} (start reached)',
    SEARCH_END_REACHED: '{!black,white!}Search torrents: {!black,yellow!}%s{!black,white!} (end reached)',
}


class SearchMode(InputKeyHandler):
    def __init__(self, torrentlist):
        super().__init__()
        self.torrentlist = torrentlist
        self.torrentview = torrentlist.torrentview
        self.search_state = SEARCH_EMPTY
        self.search_string = ''

    def update_cursor(self):
        util.safe_curs_set(util.Curser.VERY_VISIBLE)
        move_cursor(
            self.torrentlist.stdscr,
            self.torrentlist.rows - 1,
            len(self.search_string) + 17,
        )

    def set_statusbar_args(self, statusbar_args):
        statusbar_args['bottombar'] = (
            SEARCH_FORMAT[self.search_state] % self.search_string
        )
        statusbar_args['bottombar_help'] = False

    def update_colors(self, tidx, colors):
        if len(self.search_string) > 1:
            lcase_name = self.torrentview.torrent_names[tidx].lower()
            sstring_lower = self.search_string.lower()
            if lcase_name.find(sstring_lower) != -1:
                if tidx == self.torrentview.cursel:
                    pass
                elif tidx in self.torrentview.marked:
                    colors['bg'] = 'magenta'
                else:
                    colors['bg'] = 'green'
                    if colors['fg'] == 'green':
                        colors['fg'] = 'black'
                    colors['attr'] = 'bold'

    def do_search(self, direction='first'):
        """
        Performs a search on visible torrent and sets cursor to the match

        Args:
            direction (str): The direction to search. Must be one of 'first', 'last', 'next' or 'previous'

        """
        search_space = list(enumerate(self.torrentview.torrent_names))

        if direction == 'last':
            search_space = reversed(search_space)
        elif direction == 'next':
            search_space = search_space[self.torrentview.cursel + 1 :]
        elif direction == 'previous':
            search_space = reversed(search_space[: self.torrentview.cursel])

        search_string = self.search_string.lower()
        for i, n in search_space:
            n = n.lower()
            if n.find(search_string) != -1:
                self.torrentview.cursel = i
                if (
                    self.torrentview.curoff
                    + self.torrentview.torrent_rows
                    - self.torrentview.torrentlist_offset
                ) < self.torrentview.cursel:
                    self.torrentview.curoff = (
                        self.torrentview.cursel - self.torrentview.torrent_rows + 1
                    )
                elif (self.torrentview.curoff + 1) > self.torrentview.cursel:
                    self.torrentview.curoff = max(0, self.torrentview.cursel)
                self.search_state = SEARCH_SUCCESS
                return
        if direction in ['first', 'last']:
            self.search_state = SEARCH_FAILING
        elif direction == 'next':
            self.search_state = SEARCH_END_REACHED
        elif direction == 'previous':
            self.search_state = SEARCH_START_REACHED

    @overrides(InputKeyHandler)
    def handle_read(self, c):
        cname = self.torrentview.torrent_names[self.torrentview.cursel]
        refresh = True

        if c in [
            util.KEY_ESC,
            util.KEY_BELL,
        ]:  # If Escape key or CTRL-g, we abort search
            self.torrentlist.set_minor_mode(None)
            self.search_state = SEARCH_EMPTY
        elif c in [curses.KEY_BACKSPACE, util.KEY_BACKSPACE2]:
            if self.search_string:
                self.search_string = self.search_string[:-1]
                if cname.lower().find(self.search_string.lower()) != -1:
                    self.search_state = SEARCH_SUCCESS
            else:
                self.torrentlist.set_minor_mode(None)
                self.search_state = SEARCH_EMPTY
        elif c == curses.KEY_DC:
            self.search_string = ''
            self.search_state = SEARCH_SUCCESS
        elif c == curses.KEY_UP:
            self.do_search('previous')
        elif c == curses.KEY_DOWN:
            self.do_search('next')
        elif c == curses.KEY_LEFT:
            self.torrentlist.set_minor_mode(None)
            self.search_state = SEARCH_EMPTY
        elif c == ord('/'):
            self.torrentlist.set_minor_mode(None)
            self.search_state = SEARCH_EMPTY
        elif c == curses.KEY_RIGHT:
            tid = self.torrentview.current_torrent_id()
            self.torrentlist.show_torrent_details(tid)
            refresh = False
        elif c == curses.KEY_HOME:
            self.do_search('first')
        elif c == curses.KEY_END:
            self.do_search('last')
        elif c in [10, curses.KEY_ENTER]:
            self.last_mark = -1
            tid = self.torrentview.current_torrent_id()
            torrent_actions_popup(self.torrentlist, [tid], details=True)
            refresh = False
        elif c == util.KEY_ESC:
            self.search_string = ''
            self.search_state = SEARCH_EMPTY
        elif c > 31 and c < 256:
            old_search_string = self.search_string
            stroke = chr(c)
            uchar = stroke
            while not uchar:
                try:
                    uchar = stroke.decode(self.torrentlist.encoding)
                except UnicodeDecodeError:
                    c = self.torrentlist.stdscr.getch()
                    stroke += chr(c)

            if uchar:
                self.search_string += uchar

            still_matching = (
                cname.lower().find(self.search_string.lower())
                == cname.lower().find(old_search_string.lower())
                and cname.lower().find(self.search_string.lower()) != -1
            )

            if self.search_string and not still_matching:
                self.do_search()
            elif self.search_string:
                self.search_state = SEARCH_SUCCESS
        else:
            refresh = False

        if not self.search_string:
            self.search_state = SEARCH_EMPTY
            refresh = True

        if refresh:
            self.torrentlist.refresh([])

        return util.ReadState.READ



import time

import deluge.component as component
from deluge.decorators import maybe_coroutine
from deluge.ui.client import client
from deluge.ui.console.utils import colors


class EventLog(component.Component):
    """
    Prints out certain events as they are received from the core.
    """

    def __init__(self):
        component.Component.__init__(self, 'EventLog')
        self.console = component.get('ConsoleUI')
        self.prefix = '{!event!}* [%H:%M:%S] '
        self.date_change_format = 'On {!yellow!}%a, %d %b %Y{!input!} %Z:'

        event_callbacks = {
            'TorrentAddedEvent': self.on_torrent_added,
            'PreTorrentRemovedEvent': self.on_torrent_removed,
            'TorrentStateChangedEvent': self.on_torrent_state_changed,
            'TorrentFinishedEvent': self.on_torrent_finished,
            'NewVersionAvailableEvent': self.on_new_version_available,
            'SessionPausedEvent': self.on_session_paused,
            'SessionResumedEvent': self.on_session_resumed,
            'ConfigValueChangedEvent': self.on_config_value_changed,
            'PluginEnabledEvent': self.on_plugin_enabled,
            'PluginDisabledEvent': self.on_plugin_disabled,
        }

        for event, callback in event_callbacks.items():
            client.register_event_handler(event, callback)

        self.previous_time = time.localtime(0)

    @maybe_coroutine
    async def on_torrent_added(self, torrent_id, from_state):
        if from_state:
            return

        status = await client.core.get_torrent_status(torrent_id, ['name', 'state'])
        self.write(
            '{!green!}Torrent Added: {!info!}%s ({!cyan!}%s{!info!})'
            % (status['name'], torrent_id)
        )
        # Write out what state the added torrent took
        self.on_torrent_state_changed(torrent_id, status['state'])

    def on_torrent_removed(self, torrent_id):
        self.write(
            '{!red!}Torrent Removed: {!info!}%s ({!cyan!}%s{!info!})'
            % (self.console.get_torrent_name(torrent_id), torrent_id)
        )

    def on_torrent_state_changed(self, torrent_id, state):
        # It's probably a new torrent, ignore it
        if not state:
            return
        # Modify the state string color
        if state in colors.state_color:
            state = colors.state_color[state] + state

        t_name = self.console.get_torrent_name(torrent_id)

        # Again, it's most likely a new torrent
        if not t_name:
            return

        self.write(f'{state}: {{!info!}}{t_name} ({{!cyan!}}{torrent_id}{{!info!}})')

    def on_torrent_finished(self, torrent_id):
        if component.get('TorrentList').config['ring_bell']:
            import curses.beep

            curses.beep()
        self.write(
            '{!info!}Torrent Finished: %s ({!cyan!}%s{!info!})'
            % (self.console.get_torrent_name(torrent_id), torrent_id)
        )

    def on_new_version_available(self, version):
        self.write('{!input!}New Deluge version available: {!info!}%s' % (version))

    def on_session_paused(self):
        self.write('{!input!}Session Paused')

    def on_session_resumed(self):
        self.write('{!green!}Session Resumed')

    def on_config_value_changed(self, key, value):
        color = '{!white,black,bold!}'
        try:
            color = colors.type_color[type(value)]
        except KeyError:
            pass

        self.write(f'ConfigValueChanged: {{!input!}}{key}: {color}{value}')

    def write(self, s):
        current_time = time.localtime()

        date_different = False
        for field in ['tm_mday', 'tm_mon', 'tm_year']:
            c = getattr(current_time, field)
            p = getattr(self.previous_time, field)
            if c != p:
                date_different = True

        if date_different:
            string = time.strftime(self.date_change_format)
            self.console.write_event(' ')
            self.console.write_event(string)

        p = time.strftime(self.prefix)

        self.console.write_event(p + s)
        self.previous_time = current_time

    def on_plugin_enabled(self, name):
        self.write('PluginEnabled: {!info!}%s' % name)

    def on_plugin_disabled(self, name):
        self.write('PluginDisabled: {!info!}%s' % name)



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2010 Pedro Algarvio <pedro@algarvio.me>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

# The main starting point for the program. This function is called when the
# user runs the command 'deluge'.

"""Main starting point for Deluge"""

import argparse
import logging
import os
import sys

import pkg_resources

import deluge.common
import deluge.configmanager
from deluge.argparserbase import ArgParserBase
from deluge.i18n import setup_translation

DEFAULT_PREFS = {'default_ui': 'gtk'}

AMBIGUOUS_CMD_ARGS = ('-h', '--help', '-v', '-V', '--version')


def start_ui():
    """Entry point for ui script"""
    setup_translation()

    # Get the registered UI entry points
    ui_entrypoints = {}
    for entrypoint in pkg_resources.iter_entry_points('deluge.ui'):
        try:
            ui_entrypoints[entrypoint.name] = entrypoint.load()
        except ImportError:
            # Unable to load entrypoint so skip adding it.
            pass

    ui_titles = sorted(ui_entrypoints)

    def add_ui_options_group(_parser):
        """Function to enable reuse of UI Options group"""
        group = _parser.add_argument_group(_('UI Options'))
        group.add_argument(
            '-s',
            '--set-default-ui',
            dest='default_ui',
            choices=ui_titles,
            help=_('Set the default UI to be run, when no UI is specified'),
        )
        return _parser

    # Setup parser with Common Options and add UI Options group.
    parser = add_ui_options_group(ArgParserBase())

    # Parse and handle common/process group options
    options = parser.parse_known_ui_args(sys.argv, withhold=AMBIGUOUS_CMD_ARGS)

    config = deluge.configmanager.ConfigManager('ui.conf', DEFAULT_PREFS)
    log = logging.getLogger(__name__)
    log.info('Deluge ui %s', deluge.common.get_version())

    if options.default_ui:
        config['default_ui'] = options.default_ui
        config.save()
        log.info('The default UI has been changed to %s', options.default_ui)
        sys.exit(0)

    default_ui = config['default_ui']
    config.save()  # Save in case config didn't already exist.
    del config

    # We have parsed and got the config dir needed to get the default UI
    # Now create a parser for choosing the UI. We reuse the ui option group for
    # parsing to succeed and the text displayed to user, but result is not used.
    parser = add_ui_options_group(ArgParserBase(common_help=True))

    # Create subparser for each registered UI. Empty title is used to remove unwanted positional text.
    subparsers = parser.add_subparsers(
        dest='selected_ui',
        metavar='{%s} [UI args]' % ','.join(ui_titles),
        title=None,
        help=_('Alternative UI to launch, with optional ui args \n  (default UI: *)'),
    )
    for ui in ui_titles:
        parser_ui = subparsers.add_parser(
            ui,
            common_help=False,
            help=getattr(ui_entrypoints[ui], 'cmd_description', ''),
        )
        parser_ui.add_argument('ui_args', nargs=argparse.REMAINDER)
        # If the UI is set as default, indicate this in help by prefixing with a star.
        subactions = subparsers._get_subactions()
        prefix = '*' if ui == default_ui else ' '
        subactions[-1].metavar = f'{prefix} {ui}'

    # Insert a default UI subcommand unless one of the ambiguous_args are specified
    parser.set_default_subparser(default_ui, abort_opts=AMBIGUOUS_CMD_ARGS)

    # Only parse known arguments to leave the UIs to show a help message if parsing fails.
    options, remaining = parser.parse_known_args()
    selected_ui = options.selected_ui
    ui_args = remaining + options.ui_args

    # Remove the UI argument before launching the UI.
    sys.argv.remove(selected_ui)

    try:
        ui = ui_entrypoints[selected_ui](
            prog=f'{os.path.basename(sys.argv[0])} {selected_ui}', ui_args=ui_args
        )
    except KeyError:
        log.error(
            'Unable to find chosen UI: "%s". Please choose a different UI '
            'or use "--set-default-ui" to change default UI.',
            selected_ui,
        )
    except ImportError as ex:
        import traceback

        error_type, error_value, tb = sys.exc_info()
        stack = traceback.extract_tb(tb)
        last_frame = stack[-1]
        if last_frame[0] == __file__:
            log.error(
                'Unable to find chosen UI: "%s". Please choose a different UI '
                'or use "--set-default-ui" to change default UI.',
                selected_ui,
            )
        else:
            log.exception(ex)
            log.error('Encountered an error launching the request UI: %s', selected_ui)
        sys.exit(1)
    else:
        ui.start()



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge.decorators import overrides
from deluge.ui.console.modes.basemode import InputKeyHandler, move_cursor
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.widgets.fields import (
    CheckedInput,
    CheckedPlusInput,
    ComboInput,
    DividerField,
    FloatSpinInput,
    Header,
    InfoField,
    IntSpinInput,
    NoInputField,
    SelectInput,
    TextArea,
    TextField,
    TextInput,
)

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


class BaseInputPane(InputKeyHandler):
    def __init__(
        self,
        mode,
        allow_rearrange=False,
        immediate_action=False,
        set_first_input_active=True,
        border_off_west=0,
        border_off_north=0,
        border_off_east=0,
        border_off_south=0,
        active_wrap=False,
        **kwargs,
    ):
        InputKeyHandler.__init__(self)
        self.inputs = []
        self.mode = mode
        self.active_input = 0
        self.set_first_input_active = set_first_input_active
        self.allow_rearrange = allow_rearrange
        self.immediate_action = immediate_action
        self.move_active_many = 4
        self.active_wrap = active_wrap
        self.lineoff = 0
        self.border_off_west = border_off_west
        self.border_off_north = border_off_north
        self.border_off_east = border_off_east
        self.border_off_south = border_off_south
        self.last_lineoff_move = 0

        if not hasattr(self, 'visible_content_pane_height'):
            log.error(
                'The class "%s" does not have the attribute "%s" required by super class "%s"',
                self.__class__.__name__,
                'visible_content_pane_height',
                BaseInputPane.__name__,
            )
            raise AttributeError('visible_content_pane_height')

    @property
    def visible_content_pane_width(self):
        return self.mode.width

    def add_spaces(self, num):
        string = ''
        for i in range(num):
            string += '\n'

        self.add_text_area('space %d' % len(self.inputs), string)

    def add_text(self, string):
        self.add_text_area('', string)

    def move(self, r, c):
        self._cursor_row = r
        self._cursor_col = c

    def get_input(self, name):
        for e in self.inputs:
            if e.name == name:
                return e

    def _add_input(self, input_element):
        for e in self.inputs:
            if isinstance(e, NoInputField):
                continue
            if e.name == input_element.name:
                import traceback

                log.warning(
                    'Input element with name "%s" already exists in input pane (%s):\n%s',
                    input_element.name,
                    e,
                    ''.join(traceback.format_stack(limit=5)),
                )
                return

        self.inputs.append(input_element)
        if self.set_first_input_active and input_element.selectable():
            self.active_input = len(self.inputs) - 1
            self.set_first_input_active = False
        return input_element

    def add_header(self, header, space_above=False, space_below=False, **kwargs):
        return self._add_input(Header(self, header, space_above, space_below, **kwargs))

    def add_info_field(self, name, label, value):
        return self._add_input(InfoField(self, name, label, value))

    def add_text_field(self, name, message, selectable=True, col='+1', **kwargs):
        return self._add_input(
            TextField(self, name, message, selectable=selectable, col=col, **kwargs)
        )

    def add_text_area(self, name, message, **kwargs):
        return self._add_input(TextArea(self, name, message, **kwargs))

    def add_divider_field(self, name, message, **kwargs):
        return self._add_input(DividerField(self, name, message, **kwargs))

    def add_text_input(self, name, message, value='', col='+1', **kwargs):
        """
        Add a text input field

        :param message: string to display above the input field
        :param name: name of the field, for the return callback
        :param value: initial value of the field
        :param complete: should completion be run when tab is hit and this field is active
        """
        return self._add_input(
            TextInput(
                self,
                name,
                message,
                self.move,
                self.visible_content_pane_width,
                value,
                col=col,
                **kwargs,
            )
        )

    def add_select_input(self, name, message, opts, vals, default_index=0, **kwargs):
        return self._add_input(
            SelectInput(self, name, message, opts, vals, default_index, **kwargs)
        )

    def add_checked_input(self, name, message, checked=False, col='+1', **kwargs):
        return self._add_input(
            CheckedInput(self, name, message, checked=checked, col=col, **kwargs)
        )

    def add_checkedplus_input(
        self, name, message, child, checked=False, col='+1', **kwargs
    ):
        return self._add_input(
            CheckedPlusInput(
                self, name, message, child, checked=checked, col=col, **kwargs
            )
        )

    def add_float_spin_input(self, name, message, value=0.0, col='+1', **kwargs):
        return self._add_input(
            FloatSpinInput(self, name, message, self.move, value, col=col, **kwargs)
        )

    def add_int_spin_input(self, name, message, value=0, col='+1', **kwargs):
        return self._add_input(
            IntSpinInput(self, name, message, self.move, value, col=col, **kwargs)
        )

    def add_combo_input(self, name, message, choices, col='+1', **kwargs):
        return self._add_input(
            ComboInput(self, name, message, choices, col=col, **kwargs)
        )

    @overrides(InputKeyHandler)
    def handle_read(self, c):
        if not self.inputs:  # no inputs added yet
            return util.ReadState.IGNORED
        ret = self.inputs[self.active_input].handle_read(c)
        if ret != util.ReadState.IGNORED:
            if self.immediate_action:
                self.immediate_action_cb(
                    state_changed=False if ret == util.ReadState.READ else True
                )
            return ret

        ret = util.ReadState.READ

        if c == curses.KEY_UP:
            self.move_active_up(1)
        elif c == curses.KEY_DOWN:
            self.move_active_down(1)
        elif c == curses.KEY_HOME:
            self.move_active_up(len(self.inputs))
        elif c == curses.KEY_END:
            self.move_active_down(len(self.inputs))
        elif c == curses.KEY_PPAGE:
            self.move_active_up(self.move_active_many)
        elif c == curses.KEY_NPAGE:
            self.move_active_down(self.move_active_many)
        elif c == util.KEY_ALT_AND_ARROW_UP:
            self.lineoff = max(self.lineoff - 1, 0)
        elif c == util.KEY_ALT_AND_ARROW_DOWN:
            tot_height = self.get_content_height()
            self.lineoff = min(
                self.lineoff + 1, tot_height - self.visible_content_pane_height
            )
        elif c == util.KEY_CTRL_AND_ARROW_UP:
            if not self.allow_rearrange:
                return ret
            val = self.inputs.pop(self.active_input)
            self.active_input -= 1
            self.inputs.insert(self.active_input, val)
            if self.immediate_action:
                self.immediate_action_cb(state_changed=True)
        elif c == util.KEY_CTRL_AND_ARROW_DOWN:
            if not self.allow_rearrange:
                return ret
            val = self.inputs.pop(self.active_input)
            self.active_input += 1
            self.inputs.insert(self.active_input, val)
            if self.immediate_action:
                self.immediate_action_cb(state_changed=True)
        else:
            ret = util.ReadState.IGNORED
        return ret

    def get_values(self):
        vals = {}
        for i, ipt in enumerate(self.inputs):
            if not ipt.has_input():
                continue
            vals[ipt.name] = {
                'value': ipt.get_value(),
                'order': i,
                'active': self.active_input == i,
            }
        return vals

    def immediate_action_cb(self, state_changed=True):
        pass

    def move_active(self, direction, amount):
        """
        direction == -1: Up
        direction ==  1: Down

        """
        self.last_lineoff_move = direction * amount

        if direction > 0:
            if self.active_wrap:
                limit = self.active_input - 1
                if limit < 0:
                    limit = len(self.inputs) + limit
            else:
                limit = len(self.inputs) - 1
        else:
            limit = 0
            if self.active_wrap:
                limit = self.active_input + 1

        def next_move(nc, direction, limit):
            next_index = nc
            while next_index != limit:
                next_index += direction
                if direction > 0:
                    next_index %= len(self.inputs)
                elif next_index < 0:
                    next_index = len(self.inputs) + next_index

                if self.inputs[next_index].selectable():
                    return next_index
                if next_index == limit:
                    return nc
            return nc

        next_sel = self.active_input
        for a in range(amount):
            cur_sel = next_sel
            next_sel = next_move(next_sel, direction, limit)
            if cur_sel == next_sel:
                tot_height = (
                    self.get_content_height()
                    + self.border_off_north
                    + self.border_off_south
                )
                if direction > 0:
                    self.lineoff = min(
                        self.lineoff + 1, tot_height - self.visible_content_pane_height
                    )
                else:
                    self.lineoff = max(self.lineoff - 1, 0)

        if next_sel is not None:
            self.active_input = next_sel

    def move_active_up(self, amount):
        self.move_active(-1, amount)
        if self.immediate_action:
            self.immediate_action_cb(state_changed=False)

    def move_active_down(self, amount):
        self.move_active(1, amount)
        if self.immediate_action:
            self.immediate_action_cb(state_changed=False)

    def get_content_height(self):
        height = 0
        for i, ipt in enumerate(self.inputs):
            if ipt.depend_skip():
                continue
            height += ipt.height
        return height

    def ensure_active_visible(self):
        start_row = 0
        end_row = self.border_off_north
        for i, ipt in enumerate(self.inputs):
            if ipt.depend_skip():
                continue
            start_row = end_row
            end_row += ipt.height
            if i != self.active_input or not ipt.has_input():
                continue
            height = self.visible_content_pane_height
            if end_row > height + self.lineoff:
                self.lineoff += end_row - (
                    height + self.lineoff
                )  # Correct result depends on paranthesis
            elif start_row < self.lineoff:
                self.lineoff -= self.lineoff - start_row
            break

    def render_inputs(self, focused=False):
        self._cursor_row = -1
        self._cursor_col = -1
        util.safe_curs_set(util.Curser.INVISIBLE)

        self.ensure_active_visible()

        crow = self.border_off_north
        for i, ipt in enumerate(self.inputs):
            if ipt.depend_skip():
                continue
            col = self.border_off_west
            field_width = self.width - self.border_off_east - self.border_off_west
            cursor_offset = self.border_off_west

            if ipt.default_col != -1:
                default_col = int(ipt.default_col)
                if isinstance(ipt.default_col, ''.__class__) and ipt.default_col[0] in [
                    '+',
                    '-',
                ]:
                    col += default_col
                    cursor_offset += default_col
                    field_width -= default_col  # Increase to col must be reflected here
                else:
                    col = default_col
            crow += ipt.render(
                self.screen,
                crow,
                width=field_width,
                active=i == self.active_input,
                focused=focused,
                col=col,
                cursor_offset=cursor_offset,
            )

        if self._cursor_row >= 0:
            util.safe_curs_set(util.Curser.VERY_VISIBLE)
            move_cursor(self.screen, self._cursor_row, self._cursor_col)



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

try:
    import curses
except ImportError:
    pass

KEY_BELL = 7  # CTRL-/ ^G (curses.keyname(KEY_BELL) == "^G")
KEY_TAB = 9
KEY_ENTER2 = 10
KEY_ESC = 27
KEY_SPACE = 32
KEY_BACKSPACE2 = 127

KEY_ALT_AND_ARROW_UP = 564
KEY_ALT_AND_ARROW_DOWN = 523

KEY_ALT_AND_KEY_PPAGE = 553
KEY_ALT_AND_KEY_NPAGE = 548

KEY_CTRL_AND_ARROW_UP = 566
KEY_CTRL_AND_ARROW_DOWN = 525


def is_printable_chr(c):
    return c >= 32 and c <= 126


def is_int_chr(c):
    return c > 47 and c < 58


class Curser:
    INVISIBLE = 0
    NORMAL = 1
    VERY_VISIBLE = 2


def safe_curs_set(visibility):
    """
    Args:
        visibility(int): 0, 1, or 2, for invisible, normal, or very visible

    curses.curs_set fails on monochrome terminals so use this
    to ignore errors
    """
    try:
        curses.curs_set(visibility)
    except curses.error:
        pass


class ReadState:
    IGNORED = 0
    READ = 1
    CHANGED = 2



#
# Copyright (C) 2012 Bro <bro.development@gmail.com>
# Copyright (C) 2018 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import struct
import zlib

import rencode
from twisted.internet.protocol import Protocol

log = logging.getLogger(__name__)

PROTOCOL_VERSION = 1
MESSAGE_HEADER_FORMAT = '!BI'
MESSAGE_HEADER_SIZE = struct.calcsize(MESSAGE_HEADER_FORMAT)


class DelugeTransferProtocol(Protocol):
    """
    Deluge RPC wire protocol.

    Data messages are transferred with a header containing a protocol version
    and the length of the data to be transferred (payload).

    The format is::

            ubyte    uint4     bytestring
        |.version.|..size..|.....body.....|

    The version is an unsigned byte that indicates the protocol version.
    The size is a unsigned 32-bit integer that is equal to the length of the body bytestring.
    The body is the compressed rencoded byte string of the data object.
    """

    def __init__(self):
        self._buffer = b''  # TODO: Look into using bytearray instead of byte string.
        self._message_length = 0
        self._bytes_received = 0
        self._bytes_sent = 0

    def transfer_message(self, data):
        """
        Transfer the data.

        :param data: data to be transferred in a data structure serializable by rencode.
        """
        body = zlib.compress(rencode.dumps(data))
        body_len = len(body)
        message = struct.pack(
            f'{MESSAGE_HEADER_FORMAT}{body_len}s',
            PROTOCOL_VERSION,
            body_len,
            body,
        )
        self._bytes_sent += len(message)
        self.transport.write(message)

    def dataReceived(self, data):  # NOQA: N802
        """
        This method is called whenever data is received.

        :param data: a message as transferred by transfer_message, or a part of such
                     a message.

        Global variables:
            _buffer         - contains the data received
            _message_length - the length of the payload of the current message.

        """
        self._buffer += data
        self._bytes_received += len(data)

        while len(self._buffer) >= MESSAGE_HEADER_SIZE:
            if self._message_length == 0:
                self._handle_new_message()
            # We have a complete packet
            if len(self._buffer) >= self._message_length:
                self._handle_complete_message(self._buffer[: self._message_length])
                # Remove message data from buffer
                self._buffer = self._buffer[self._message_length :]
                self._message_length = 0
            else:
                break

    def _handle_new_message(self):
        """
        Handle the start of a new message. This method is called only when the
        beginning of the buffer contains data from a new message (i.e. the header).

        """
        try:
            # Read the first bytes of the message (MESSAGE_HEADER_SIZE bytes)
            header = self._buffer[:MESSAGE_HEADER_SIZE]
            # Extract the length stored as an unsigned 32-bit integer
            version, self._message_length = struct.unpack(MESSAGE_HEADER_FORMAT, header)
            if version != PROTOCOL_VERSION:
                raise Exception(
                    'Received invalid protocol version: {}. PROTOCOL_VERSION is {}.'.format(
                        version, PROTOCOL_VERSION
                    )
                )
            # Remove the header from the buffer
            self._buffer = self._buffer[MESSAGE_HEADER_SIZE:]
        except Exception as ex:
            log.warning('Error occurred when parsing message header: %s.', ex)
            log.warning(
                'This version of Deluge cannot communicate with the sender of this data.'
            )
            self._message_length = 0
            self._buffer = b''

    def _handle_complete_message(self, data):
        """
        Handles a complete message as it is transferred on the network.

        :param data: a zlib compressed string encoded with rencode.

        """
        try:
            self.message_received(
                rencode.loads(zlib.decompress(data), decode_utf8=True)
            )
        except Exception as ex:
            log.warning(
                'Failed to decompress (%d bytes) and load serialized data with rencode: %s',
                len(data),
                ex,
            )

    def get_bytes_recv(self):
        """
        Returns the number of bytes received.

        :returns: the number of bytes received
        :rtype: int

        """
        return self._bytes_received

    def get_bytes_sent(self):
        """
        Returns the number of bytes sent.

        :returns: the number of bytes sent
        :rtype: int

        """
        return self._bytes_sent

    def message_received(self, message):
        """Override this method to receive the complete message"""
        pass



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import shlex

from twisted.internet import defer

from deluge.ui.client import client
from deluge.ui.console.parser import OptionParser, OptionParserError
from deluge.ui.console.utils.colors import strip_colors

log = logging.getLogger(__name__)


class Commander:
    def __init__(self, cmds, interactive=False):
        self._commands = cmds
        self.interactive = interactive

    def write(self, line):
        print(strip_colors(line))

    def do_command(self, cmd_line):
        """Run a console command.

        Args:
            cmd_line (str): Console command.

        Returns:
            Deferred: A deferred that fires when the command has been executed.

        """
        options = self.parse_command(cmd_line)
        if options:
            return self.exec_command(options)
        return defer.succeed(None)

    def exit(self, status=0, msg=None):
        self._exit = True
        if msg:
            print(msg)

    def parse_command(self, cmd_line):
        """Parse a console command and process with argparse.

        Args:
            cmd_line (str): Console command.

        Returns:
            argparse.Namespace: The parsed command.

        """
        if not cmd_line:
            return
        cmd, _, line = cmd_line.partition(' ')
        try:
            parser = self._commands[cmd].create_parser()
        except KeyError:
            self.write('{!error!}Unknown command: %s' % cmd)
            return

        try:
            args = [cmd] + self._commands[cmd].split(line)
        except ValueError as ex:
            self.write('{!error!}Error parsing command: %s' % ex)
            return

        # Do a little hack here to print 'command --help' properly
        parser._print_help = parser.print_help

        def print_help(f=None):
            if self.interactive:
                self.write(parser.format_help())
            else:
                parser._print_help(f)

        parser.print_help = print_help

        # Only these commands can be run when not connected to a daemon
        not_connected_cmds = ['help', 'connect', 'quit']
        aliases = []
        for c in not_connected_cmds:
            aliases.extend(self._commands[c].aliases)
        not_connected_cmds.extend(aliases)

        if not client.connected() and cmd not in not_connected_cmds:
            self.write(
                '{!error!}Not connected to a daemon, please use the connect command first.'
            )
            return

        try:
            options = parser.parse_args(args=args)
            options.command = cmd
        except TypeError as ex:
            self.write('{!error!}Error parsing options: %s' % ex)
            import traceback

            self.write('%s' % traceback.format_exc())
            return
        except OptionParserError as ex:
            import traceback

            log.warning('Error parsing command "%s":  %s', args, ex)
            self.write('{!error!} %s' % ex)
            parser.print_help()
            return

        if getattr(parser, '_exit', False):
            return
        return options

    def exec_command(self, options, *args):
        """Execute a console command.

        Args:
            options (argparse.Namespace): The command to execute.

        Returns:
            Deferred: A deferred that fires when command has been executed.

        """
        try:
            ret = self._commands[options.command].handle(options)
        except Exception as ex:  # pylint: disable=broad-except
            self.write('{!error!} %s' % ex)
            log.exception(ex)
            import traceback

            self.write('%s' % traceback.format_exc())
            return defer.succeed(True)
        else:
            return ret


class BaseCommand:
    usage = None
    interactive_only = False
    aliases = []
    _name = 'base'
    epilog = ''

    def complete(self, text, *args):
        return []

    def handle(self, options):
        pass

    @property
    def name(self):
        return self._name

    @property
    def name_with_alias(self):
        return '/'.join([self._name] + self.aliases)

    @property
    def description(self):
        return self.__doc__

    def split(self, text):
        text = text.replace('\\', '\\\\')
        result = shlex.split(text)
        for i, s in enumerate(result):
            result[i] = s.replace(r'\ ', ' ')
        result = [s for s in result if s != '']
        return result

    def create_parser(self):
        opts = {
            'prog': self.name_with_alias,
            'description': self.__doc__,
            'epilog': self.epilog,
        }
        if self.usage:
            opts['usage'] = self.usage
        parser = OptionParser(**opts)
        parser.add_argument(self.name, metavar='')
        parser.base_parser = parser
        self.add_arguments(parser)
        return parser

    def add_subparser(self, subparsers):
        opts = {
            'prog': self.name_with_alias,
            'help': self.__doc__,
            'description': self.__doc__,
        }
        if self.usage:
            opts['usage'] = self.usage

        # A workaround for aliases showing as duplicate command names in help output.
        for cmd_name in sorted([self.name] + self.aliases):
            if cmd_name not in subparsers._name_parser_map:
                if cmd_name in self.aliases:
                    opts['help'] = _('`%s` alias') % self.name
                parser = subparsers.add_parser(cmd_name, **opts)
                break

        self.add_arguments(parser)

    def add_arguments(self, parser):
        pass



#
# Original file from BitTorrent-5.3-GPL.tar.gz
# Copyright (C) Bram Cohen
#
# Modifications for use in Deluge:
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import copy
import logging
import os.path
import time
from enum import Enum
from hashlib import sha1 as sha
from hashlib import sha256

import deluge.component as component
from deluge.bencode import bencode
from deluge.common import utf8_encode_structure
from deluge.event import CreateTorrentProgressEvent

log = logging.getLogger(__name__)

ignore = ['core', 'CVS', 'Thumbs.db', 'desktop.ini']

noncharacter_translate = {}
for i in range(0xD800, 0xE000):
    noncharacter_translate[i] = ord('-')
for i in range(0xFDD0, 0xFDF0):
    noncharacter_translate[i] = ord('-')
for i in (0xFFFE, 0xFFFF):
    noncharacter_translate[i] = ord('-')


def gmtime():
    return time.mktime(time.gmtime())


def dummy(*v):
    pass


class TorrentFormat(str, Enum):
    V1 = 'v1'
    V2 = 'v2'
    HYBRID = 'hybrid'

    @classmethod
    def _missing_(cls, value):
        if not value:
            return None

        value = value.lower()
        for member in cls:
            if member.value == value:
                return member

    def to_lt_flag(self):
        if self.value == 'v1':
            return 64
        if self.value == 'v2':
            return 32
        return 0

    def includes_v1(self):
        return self == self.__class__.V1 or self == self.__class__.HYBRID

    def includes_v2(self):
        return self == self.__class__.V2 or self == self.__class__.HYBRID


class RemoteFileProgress:
    def __init__(self, session_id):
        self.session_id = session_id

    def __call__(self, piece_count, num_pieces):
        component.get('RPCServer').emit_event_for_session_id(
            self.session_id, CreateTorrentProgressEvent(piece_count, num_pieces)
        )


def make_meta_file_content(
    path,
    url,
    piece_length,
    progress=None,
    title=None,
    comment=None,
    safe=None,
    content_type=None,
    webseeds=None,
    name=None,
    private=False,
    created_by=None,
    trackers=None,
    torrent_format=TorrentFormat.V1,
):
    data = {'creation date': int(gmtime())}
    if url:
        data['announce'] = url.strip()

    if progress is None:
        progress = dummy
        try:
            session_id = component.get('RPCServer').get_session_id()
        except KeyError:
            pass
        else:
            if session_id:
                progress = RemoteFileProgress(session_id)

    info, piece_layers = makeinfo(
        path,
        piece_length,
        progress,
        name,
        content_type,
        private,
        torrent_format,
    )

    # check_info(info)
    data['info'] = info
    if piece_layers is not None:
        data['piece layers'] = piece_layers
    if title:
        data['title'] = title.encode('utf8')
    if comment:
        data['comment'] = comment.encode('utf8')
    if safe:
        data['safe'] = safe.encode('utf8')

    httpseeds = []
    url_list = []

    if webseeds:
        for webseed in webseeds:
            if webseed.endswith('.php'):
                httpseeds.append(webseed)
            else:
                url_list.append(webseed)

    if url_list:
        data['url-list'] = url_list
    if httpseeds:
        data['httpseeds'] = httpseeds
    if created_by:
        data['created by'] = created_by.encode('utf8')

    if trackers and (len(trackers[0]) > 1 or len(trackers) > 1):
        data['announce-list'] = trackers

    data['encoding'] = 'UTF-8'
    return bencode(utf8_encode_structure(data))


def default_meta_file_path(content_path):
    a, b = os.path.split(content_path)
    if b == '':
        f = a + '.torrent'
    else:
        f = os.path.join(a, b + '.torrent')
    return f


def make_meta_file(
    path,
    url,
    piece_length,
    progress=None,
    title=None,
    comment=None,
    safe=None,
    content_type=None,
    target=None,
    webseeds=None,
    name=None,
    private=False,
    created_by=None,
    trackers=None,
):
    if not target:
        target = default_meta_file_path(path)

    file_content = make_meta_file_content(
        path,
        url,
        piece_length,
        progress=progress,
        title=title,
        comment=comment,
        safe=safe,
        content_type=content_type,
        webseeds=webseeds,
        name=name,
        private=private,
        created_by=created_by,
        trackers=trackers,
    )

    with open(target, 'wb') as file_:
        file_.write(file_content)


def calcsize(path):
    total = 0
    for s in subfiles(os.path.abspath(path)):
        total += os.path.getsize(s[1])
    return total


def _next_pow2(num):
    import math

    if not num:
        return 1
    return 2 ** math.ceil(math.log2(num))


def _sha256_merkle_root(leafs, nb_leafs, padding, in_place=True) -> bytes:
    """
    Build the root of the merkle hash tree from the (possibly incomplete) leafs layer.
    If len(leafs) < nb_leafs, it will be padded with the padding repeated as many times
    as needed to have nb_leafs in total.
    """
    if not in_place:
        leafs = copy.copy(leafs)

    while nb_leafs > 1:
        nb_leafs = nb_leafs // 2
        for i in range(nb_leafs):
            node1 = leafs[2 * i] if 2 * i < len(leafs) else padding
            node2 = leafs[2 * i + 1] if 2 * i + 1 < len(leafs) else padding
            h = sha256(node1)
            h.update(node2)
            if i < len(leafs):
                leafs[i] = h.digest()
            else:
                leafs.append(h.digest())
    return leafs[0] if leafs else padding


def _sha256_buffer_blocks(buffer, block_len):
    import math

    nb_blocks = math.ceil(len(buffer) / block_len)
    blocks = [
        sha256(buffer[i * block_len : (i + 1) * block_len]).digest()
        for i in range(nb_blocks)
    ]
    return blocks


def makeinfo_lt(
    path, piece_length, name=None, private=False, torrent_format=TorrentFormat.V1
):
    """
    Make info using via the libtorrent library.
    """
    from deluge._libtorrent import lt

    if not name:
        name = os.path.split(path)[1]

    fs = lt.file_storage()
    if os.path.isfile(path):
        lt.add_files(fs, path)
    else:
        for p, f in subfiles(path):
            fs.add_file(os.path.join(name, *p), os.path.getsize(f))
    torrent = lt.create_torrent(
        fs, piece_size=piece_length, flags=torrent_format.to_lt_flag()
    )

    lt.set_piece_hashes(torrent, os.path.dirname(path))
    torrent.set_priv(private)

    t = torrent.generate()
    info = t[b'info']
    pieces_layers = t.get(b'piece layers', None)

    return info, pieces_layers


def makeinfo(
    path,
    piece_length,
    progress,
    name=None,
    content_type=None,
    private=False,
    torrent_format=TorrentFormat.V1,
):
    # HEREDAVE. If path is directory, how do we assign content type?

    v2_block_len = 2**14  # 16 KiB
    v2_blocks_per_piece = 1
    v2_block_padding = b''
    v2_piece_padding = b''
    if torrent_format.includes_v2():
        if _next_pow2(piece_length) != piece_length or piece_length < v2_block_len:
            raise ValueError(
                'Bittorrent v2 piece size must be a power of 2; and bigger than 16 KiB'
            )

        v2_blocks_per_piece = piece_length // v2_block_len
        v2_block_padding = bytes(32)  # 32 = size of sha256 in bytes
        v2_piece_padding = _sha256_merkle_root(
            [], nb_leafs=v2_blocks_per_piece, padding=v2_block_padding
        )

    path = os.path.abspath(path)
    files = []
    pieces = []
    file_tree = {}
    piece_layers = {}
    if os.path.isdir(path):
        if not name:
            name = os.path.split(path)[1]
        subs = subfiles(path)
        if torrent_format.includes_v2():
            subs = sorted(subs)
        length = None
        totalsize = 0.0
        for p, f in subs:
            totalsize += os.path.getsize(f)
    else:
        name = os.path.split(path)[1]
        subs = [([name], path)]
        length = os.path.getsize(path)
        totalsize = length
    is_multi_file = len(subs) > 1
    sh = sha()
    done = 0
    totalhashed = 0

    next_progress_event = piece_length
    for p, f in subs:
        file_pieces_v2 = []
        pos = 0
        size = os.path.getsize(f)
        p2 = [n.encode('utf8') for n in p]
        if content_type:
            files.append(
                {b'length': size, b'path': p2, b'content_type': content_type}
            )  # HEREDAVE. bad for batch!
        else:
            files.append({b'length': size, b'path': p2})
        with open(f, 'rb') as file_:
            while pos < size:
                to_read = min(size - pos, piece_length)
                buffer = memoryview(file_.read(to_read))
                pos += to_read

                if torrent_format.includes_v1():
                    a = piece_length - done
                    for sub_buffer in (buffer[:a], buffer[a:]):
                        if sub_buffer:
                            sh.update(sub_buffer)
                            done += len(sub_buffer)

                            if done == piece_length:
                                pieces.append(sh.digest())
                                done = 0
                                sh = sha()
                if torrent_format.includes_v2():
                    block_hashes = _sha256_buffer_blocks(buffer, v2_block_len)
                    num_leafs = v2_blocks_per_piece
                    if size <= piece_length:
                        # The special case when the file is smaller than a piece: only pad till the next power of 2
                        num_leafs = _next_pow2(len(block_hashes))
                    root = _sha256_merkle_root(
                        block_hashes, num_leafs, v2_block_padding, in_place=True
                    )
                    file_pieces_v2.append(root)

                totalhashed += to_read
                if totalhashed >= next_progress_event:
                    next_progress_event = totalhashed + piece_length
                    progress(totalhashed, totalsize)

        if torrent_format == TorrentFormat.HYBRID and is_multi_file and done > 0:
            # Add padding file to force piece-alignment
            padding = piece_length - done
            sh.update(bytes(padding))
            files.append(
                {
                    b'length': padding,
                    b'attr': b'p',
                    b'path': [b'.pad', str(padding).encode()],
                }
            )
            pieces.append(sh.digest())
            done = 0
            sh = sha()

        if torrent_format.includes_v2():
            # add file to the `file tree` and, if needed, to the `piece layers` structures
            pieces_root = _sha256_merkle_root(
                file_pieces_v2,
                _next_pow2(len(file_pieces_v2)),
                v2_piece_padding,
                in_place=False,
            )
            dst_directory = file_tree
            for directory in p2[:-1]:
                dst_directory = dst_directory.setdefault(directory, {})
            dst_directory[p2[-1]] = {
                b'': {
                    b'length': size,
                    b'pieces root': pieces_root,
                }
            }
            if len(file_pieces_v2) > 1:
                piece_layers[pieces_root] = b''.join(file_pieces_v2)

    if done > 0:
        pieces.append(sh.digest())
    progress(totalsize, totalsize)

    info = {
        b'piece length': piece_length,
        b'name': name.encode('utf8'),
    }
    if private:
        info[b'private'] = 1
    if content_type:
        info[b'content_type'] = content_type
    if torrent_format.includes_v1():
        info[b'pieces'] = b''.join(pieces)
        if is_multi_file:
            info[b'files'] = files
        else:
            info[b'length'] = length
    if torrent_format.includes_v2():
        info.update(
            {
                b'meta version': 2,
                b'file tree': file_tree,
            }
        )
    return info, piece_layers if torrent_format.includes_v2() else None


def subfiles(d):
    r = []
    stack = [([], d)]
    while stack:
        p, n = stack.pop()
        if os.path.isdir(n):
            for s in os.listdir(n):
                if s not in ignore and not s.startswith('.'):
                    stack.append((p + [s], os.path.join(n, s)))
        else:
            r.append((p, n))
    return r



#
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os

from deluge import component
from deluge.configmanager import ConfigManager
from deluge.pluginmanagerbase import PluginManagerBase
from deluge.ui.client import client

log = logging.getLogger(__name__)


def gather_info(plugin):
    # Get the scripts for the plugin
    scripts = getattr(plugin, 'scripts', ())
    debug_scripts = getattr(plugin, 'debug_scripts') or scripts

    directories = []
    for script in scripts + debug_scripts:
        if os.path.dirname(script) not in directories:
            directories.append(os.path.dirname(script))

    return {
        'scripts': scripts,
        'debug_scripts': debug_scripts,
        'script_directories': directories,
    }


class PluginManager(PluginManagerBase, component.Component):
    def __init__(self):
        component.Component.__init__(self, 'Web.PluginManager')
        self.config = ConfigManager('web.conf')
        PluginManagerBase.__init__(self, 'web.conf', 'deluge.plugin.web')

        client.register_event_handler(
            'PluginEnabledEvent', self._on_plugin_enabled_event
        )
        client.register_event_handler(
            'PluginDisabledEvent', self._on_plugin_disabled_event
        )

    def _on_get_enabled_plugins(self, plugins):
        for plugin in plugins:
            self.enable_plugin(plugin)

    def _on_plugin_enabled_event(self, name):
        self.enable_plugin(name)

    def _on_plugin_disabled_event(self, name):
        self.disable_plugin(name)

    def disable_plugin(self, name):
        # Get the plugin instance
        try:
            plugin = component.get('WebPlugin.' + name)
        except KeyError:
            log.debug(
                '%s plugin contains no WebUI code, ignoring WebUI disable call.', name
            )
            return

        info = gather_info(plugin)

        scripts = component.get('Scripts')
        for script in info['scripts']:
            scripts.remove_script(f'{name.lower()}/{os.path.basename(script).lower()}')

        for script in info['debug_scripts']:
            scripts.remove_script(
                f'{name.lower()}/{os.path.basename(script).lower()}', 'debug'
            )
            scripts.remove_script(
                f'{name.lower()}/{os.path.basename(script).lower()}', 'dev'
            )

        super().disable_plugin(name)

    def enable_plugin(self, name):
        super().enable_plugin(name)

        # Get the plugin instance
        try:
            plugin = component.get('WebPlugin.' + name)
        except KeyError:
            log.info(
                '%s plugin contains no WebUI code, ignoring WebUI enable call.', name
            )
            return

        info = gather_info(plugin)

        scripts = component.get('Scripts')
        for script in info['scripts']:
            log.debug('adding script %s for %s', name, os.path.basename(script))
            scripts.add_script(f'{name.lower()}/{os.path.basename(script)}', script)

        for script in info['debug_scripts']:
            log.debug('adding debug script %s for %s', name, os.path.basename(script))
            scripts.add_script(
                f'{name.lower()}/{os.path.basename(script)}', script, 'debug'
            )
            scripts.add_script(
                f'{name.lower()}/{os.path.basename(script)}', script, 'dev'
            )

    def start(self):
        """
        Start up the plugin manager
        """
        # Update the enabled plugins from the core
        d = client.core.get_enabled_plugins()
        d.addCallback(self._on_get_enabled_plugins)

    def stop(self):
        """
        Stop the plugin manager
        """
        self.disable_plugins()
        client.deregister_event_handler(
            'PluginEnabledEvent', self._on_plugin_enabled_event
        )
        client.deregister_event_handler(
            'PluginDisabledEvent', self._on_plugin_disabled_event
        )

    def update(self):
        pass

    def get_plugin_resources(self, name):
        # Get the plugin instance
        try:
            plugin = component.get('WebPlugin.' + name)
        except KeyError:
            log.info('Plugin has no web ui')
            return
        info = gather_info(plugin)
        info['name'] = name
        info['scripts'] = [
            f'js/{name.lower()}/{os.path.basename(s)}' for s in info['scripts']
        ]
        info['debug_scripts'] = [
            f'js/{name.lower()}/{os.path.basename(s)}' for s in info['debug_scripts']
        ]
        del info['script_directories']
        return info



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""The Deluge daemon"""

import logging
import os
import socket

from twisted.internet import reactor

import deluge.component as component
from deluge.common import get_version, is_ip, is_process_running, windows_check
from deluge.configmanager import get_config_dir
from deluge.core.core import Core
from deluge.core.rpcserver import RPCServer, export
from deluge.error import DaemonRunningError

if windows_check():
    from win32api import SetConsoleCtrlHandler
    from win32con import CTRL_CLOSE_EVENT, CTRL_SHUTDOWN_EVENT

log = logging.getLogger(__name__)


def is_daemon_running(pid_file):
    """
    Check for another running instance of the daemon using the same pid file.

    Args:
        pid_file: The location of the file with pid, port values.

    Returns:
        bool: True is daemon is running, False otherwise.

    """

    try:
        with open(pid_file) as _file:
            pid, port = (int(x) for x in _file.readline().strip().split(';'))
    except (OSError, ValueError):
        return False

    if is_process_running(pid):
        # Ensure it's a deluged process by trying to open a socket to it's port.
        _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            _socket.connect(('127.0.0.1', port))
        except OSError:
            # Can't connect, so pid is not a deluged process.
            return False
        else:
            # This is a deluged process!
            _socket.close()
            return True


class Daemon:
    """The Deluge Daemon class"""

    def __init__(
        self,
        listen_interface=None,
        outgoing_interface=None,
        interface=None,
        port=None,
        standalone=False,
        read_only_config_keys=None,
    ):
        """
        Args:
            listen_interface (str, optional): The IP address to listen to
                BitTorrent connections on.
            outgoing_interface (str, optional): The network interface name or
                IP address to open outgoing BitTorrent connections on.
            interface (str, optional): The IP address the daemon will
                listen for UI connections on.
            port (int, optional): The port the daemon will listen for UI
                connections on.
            standalone (bool, optional): If True the client is in Standalone
                mode otherwise, if False, start the daemon as separate process.
            read_only_config_keys (list of str, optional): A list of config
                keys that will not be altered by core.set_config() RPC method.
        """
        self.standalone = standalone
        self.pid_file = get_config_dir('deluged.pid')
        log.info('Deluge daemon %s', get_version())
        if is_daemon_running(self.pid_file):
            raise DaemonRunningError(
                'Deluge daemon already running with this config directory!'
            )

        # Twisted catches signals to terminate, so just have it call the shutdown method.
        reactor.addSystemEventTrigger('before', 'shutdown', self._shutdown)

        # Catch some Windows specific signals
        if windows_check():

            def win_handler(ctrl_type):
                """Handle the Windows shutdown or close events."""
                log.debug('windows handler ctrl_type: %s', ctrl_type)
                if ctrl_type == CTRL_CLOSE_EVENT or ctrl_type == CTRL_SHUTDOWN_EVENT:
                    self._shutdown()
                    return 1

            SetConsoleCtrlHandler(win_handler)

        # Start the core as a thread and join it until it's done
        self.core = Core(
            listen_interface=listen_interface,
            outgoing_interface=outgoing_interface,
            read_only_config_keys=read_only_config_keys,
        )

        if port is None:
            port = self.core.config['daemon_port']
        self.port = port

        if interface and not is_ip(interface):
            log.error('Invalid UI interface (must be IP Address): %s', interface)
            interface = None

        self.rpcserver = RPCServer(
            port=port,
            allow_remote=self.core.config['allow_remote'],
            listen=not standalone,
            interface=interface,
        )

        log.debug(
            'Listening to UI on: %s:%s and bittorrent on: %s Making connections out on: %s',
            interface,
            port,
            listen_interface,
            outgoing_interface,
        )

    def start(self):
        # Register the daemon and the core RPCs
        self.rpcserver.register_object(self.core)
        self.rpcserver.register_object(self)

        # Make sure we start the PreferencesManager first
        component.start('PreferencesManager')

        if not self.standalone:
            log.info('Deluge daemon starting...')
            # Create pid file to track if deluged is running, also includes the port number.
            pid = os.getpid()
            log.debug('Storing pid %s & port %s in: %s', pid, self.port, self.pid_file)
            with open(self.pid_file, 'w') as _file:
                _file.write(f'{pid};{self.port}\n')

            component.start()

            try:
                reactor.run()
            finally:
                log.debug('Remove pid file: %s', self.pid_file)
                os.remove(self.pid_file)
                log.info('Deluge daemon shutdown successfully')

    @export()
    def shutdown(self, *args, **kwargs):
        log.debug('Deluge daemon shutdown requested...')
        reactor.callLater(0, reactor.stop)

    def _shutdown(self, *args, **kwargs):
        log.info('Deluge daemon shutting down, waiting for components to shutdown...')
        if not self.standalone:
            return component.shutdown()

    @export()
    def get_method_list(self):
        """Returns a list of the exported methods."""
        return self.rpcserver.get_method_list()

    @export()
    def get_version(self):
        """Returns the daemon version"""
        return get_version()

    @export(1)
    def authorized_call(self, rpc):
        """Determines if session auth_level is authorized to call RPC.

        Args:
            rpc (str): A RPC, e.g. core.get_torrents_status

        Returns:
            bool: True if authorized to call RPC, otherwise False.
        """
        if rpc not in self.get_method_list():
            return False

        return (
            self.rpcserver.get_session_auth_level()
            >= self.rpcserver.get_rpc_auth_level(rpc)
        )



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import locale
import logging
import os
import sys

from twisted.internet import defer, error, reactor

import deluge.common
import deluge.component as component
from deluge.configmanager import ConfigManager
from deluge.decorators import maybe_coroutine, overrides
from deluge.ui.client import client
from deluge.ui.console.eventlog import EventLog
from deluge.ui.console.modes.addtorrents import AddTorrents
from deluge.ui.console.modes.basemode import TermResizeHandler
from deluge.ui.console.modes.cmdline import CmdLine
from deluge.ui.console.modes.eventview import EventView
from deluge.ui.console.modes.preferences import Preferences
from deluge.ui.console.modes.torrentdetail import TorrentDetail
from deluge.ui.console.modes.torrentlist.torrentlist import TorrentList
from deluge.ui.console.utils import colors
from deluge.ui.console.utils.config import migrate_1_to_2
from deluge.ui.console.widgets import StatusBars
from deluge.ui.coreconfig import CoreConfig
from deluge.ui.sessionproxy import SessionProxy

log = logging.getLogger(__name__)

DEFAULT_CONSOLE_PREFS = {
    'ring_bell': False,
    'first_run': True,
    'language': '',
    'torrentview': {
        'sort_primary': 'queue',
        'sort_secondary': 'name',
        'show_sidebar': True,
        'sidebar_width': 25,
        'separate_complete': True,
        'move_selection': True,
        'columns': {},
    },
    'addtorrents': {
        'show_misc_files': False,  # TODO: Showing/hiding this
        'show_hidden_folders': False,  # TODO: Showing/hiding this
        'sort_column': 'date',
        'reverse_sort': True,
        'last_path': '~',
    },
    'cmdline': {
        'ignore_duplicate_lines': False,
        'third_tab_lists_all': False,
        'torrents_per_tab_press': 15,
        'save_command_history': True,
    },
}


class MockConsoleLog:
    def write(self, data):
        pass

    def flush(self):
        pass


class ConsoleUI(component.Component, TermResizeHandler):
    def __init__(self, options, cmds, log_stream):
        component.Component.__init__(self, 'ConsoleUI')
        TermResizeHandler.__init__(self)
        self.options = options
        self.log_stream = log_stream

        # keep track of events for the log view
        self.events = []
        self.torrents = []
        self.statusbars = None
        self.modes = {}
        self.active_mode = None
        self.initialized = False

        try:
            locale.setlocale(locale.LC_ALL, '')
            self.encoding = locale.getpreferredencoding()
        except locale.Error:
            self.encoding = sys.getdefaultencoding()

        log.debug('Using encoding: %s', self.encoding)

        # start up the session proxy
        self.sessionproxy = SessionProxy()

        client.set_disconnect_callback(self.on_client_disconnect)

        # Set the interactive flag to indicate where we should print the output
        self.interactive = True
        self._commands = cmds
        self.coreconfig = CoreConfig()

    def start_ui(self):
        """Start the console UI.

        Note: When running console UI reactor.run() will be called which
              effectively blocks this function making the return value
              insignificant. However, when running unit tests, the reacor is
              replaced by a mock object, leaving the return deferred object
              necessary for the tests to run properly.

        Returns:
            Deferred: If valid commands are provided, a deferred that fires when
                 all commands are executed. Else None is returned.
        """
        if self.options.parsed_cmds:
            # Non-Interactive mode
            self.interactive = False
            if not self._commands:
                print('No valid console commands found')
                return

            deferred = self.exec_args(self.options)
            reactor.run()
            return deferred

        # Interactive

        # We use the curses.wrapper function to prevent the console from getting
        # messed up if an uncaught exception is experienced.
        try:
            from curses import wrapper
        except ImportError:
            wrapper = None

        if deluge.common.windows_check():
            print(
                """\nDeluge-console does not run in interactive mode on Windows. \n
Please use commands from the command line, e.g.:\n
deluge-console.exe help
deluge-console.exe info
deluge-console.exe "add --help"
deluge-console.exe "add -p c:\\mytorrents c:\\new.torrent"
"""
            )
            return

        # We don't ever want log output to terminal when running in
        # interactive mode, so insert a dummy here
        self.log_stream.out = MockConsoleLog()

        # Set Esc key delay to 0 to avoid a very annoying delay
        # due to curses waiting in case of other key are pressed
        # after ESC is pressed
        os.environ.setdefault('ESCDELAY', '0')

        wrapper(self.run)

    @maybe_coroutine
    async def quit(self):
        if client.connected():
            await client.disconnect()

        try:
            reactor.stop()
        except error.ReactorNotRunning:
            pass

    @maybe_coroutine
    async def exec_args(self, options):
        """Execute console commands from command line."""
        from deluge.ui.console.cmdline.command import Commander

        commander = Commander(self._commands)
        try:
            if not self.interactive and options.parsed_cmds[0].command == 'connect':
                await commander.exec_command(options.parsed_cmds.pop(0))
            else:
                daemon_options = (
                    options.daemon_addr,
                    options.daemon_port,
                    options.daemon_user,
                    options.daemon_pass,
                )
                log.info(
                    'Connect: host=%s, port=%s, username=%s',
                    *daemon_options[0:3],
                )
                await client.connect(*daemon_options)
        except Exception as reason:
            print(
                'Could not connect to daemon: %s:%s\n %s'
                % (options.daemon_addr, options.daemon_port, reason)
            )
            commander.do_command('quit')

        await self.start_console()
        # Wait for RPCs in start() to finish before processing commands.
        await self.started_deferred

        for cmd in options.parsed_cmds:
            if cmd.command in ('quit', 'exit'):
                break
            await commander.exec_command(cmd)

        commander.do_command('quit')

    def run(self, stdscr):
        """This method is called by the curses.wrapper to start the mainloop and screen.

        Args:
            stdscr (_curses.curses window): curses screen passed in from curses.wrapper.

        """
        # We want to do an interactive session, so start up the curses screen and
        # pass it the function that handles commands
        colors.init_colors()
        self.stdscr = stdscr
        self.config = ConfigManager(
            'console.conf', defaults=DEFAULT_CONSOLE_PREFS, file_version=2
        )
        self.config.run_converter((0, 1), 2, migrate_1_to_2)

        self.statusbars = StatusBars()
        from deluge.ui.console.modes.connectionmanager import ConnectionManager

        self.register_mode(ConnectionManager(stdscr, self.encoding), set_mode=True)

        torrentlist = self.register_mode(TorrentList(self.stdscr, self.encoding))
        self.register_mode(CmdLine(self.stdscr, self.encoding))
        self.register_mode(EventView(torrentlist, self.stdscr, self.encoding))
        self.register_mode(
            TorrentDetail(torrentlist, self.stdscr, self.config, self.encoding)
        )
        self.register_mode(
            Preferences(torrentlist, self.stdscr, self.config, self.encoding)
        )
        self.register_mode(
            AddTorrents(torrentlist, self.stdscr, self.config, self.encoding)
        )

        self.eventlog = EventLog()

        self.active_mode.topbar = (
            '{!status!}Deluge ' + deluge.common.get_version() + ' Console'
        )
        self.active_mode.bottombar = '{!status!}'
        self.active_mode.refresh()
        # Start the twisted mainloop
        reactor.run()

    @overrides(TermResizeHandler)
    def on_resize(self, *args):
        rows, cols = super().on_resize(*args)
        for mode in self.modes:
            self.modes[mode].on_resize(rows, cols)

    def register_mode(self, mode, set_mode=False):
        self.modes[mode.mode_name] = mode
        if set_mode:
            self.set_mode(mode.mode_name)
        return mode

    def set_mode(self, mode_name, refresh=False):
        log.debug('Setting console mode: %s', mode_name)
        mode = self.modes.get(mode_name, None)
        if mode is None:
            log.error('Non-existent mode requested: %s', mode_name)
            return
        self.stdscr.erase()

        if self.active_mode:
            self.active_mode.pause()
            d = component.pause([self.active_mode.mode_name])

            def on_mode_paused(result, mode, *args):
                from deluge.ui.console.widgets.popup import PopupsHandler

                if isinstance(mode, PopupsHandler):
                    if mode.popup is not None:
                        # If popups are not removed, they are still referenced in the memory
                        # which can cause issues as the popup's screen will not be destroyed.
                        # This can lead to the popup border being visible for short periods
                        # while the current modes' screen is repainted.
                        log.error(
                            'Mode "%s" still has popups available after being paused.'
                            ' Ensure all popups are removed on pause!',
                            mode.popup.title,
                        )

            d.addCallback(on_mode_paused, self.active_mode)
            reactor.removeReader(self.active_mode)

        self.active_mode = mode
        self.statusbars.screen = self.active_mode

        # The Screen object is designed to run as a twisted reader so that it
        # can use twisted's select poll for non-blocking user input.
        reactor.addReader(self.active_mode)
        self.stdscr.clear()

        if self.active_mode._component_state == 'Stopped':
            component.start([self.active_mode.mode_name])
        else:
            component.resume([self.active_mode.mode_name])

        mode.resume()
        if refresh:
            mode.refresh()
        return mode

    def switch_mode(self, func, error_smg):
        def on_stop(arg):
            if arg and True in arg[0]:
                func()
            else:
                self.messages.append(('Error', error_smg))

        component.stop(['TorrentList']).addCallback(on_stop)

    def is_active_mode(self, mode):
        return mode == self.active_mode

    @maybe_coroutine
    async def start_components(self):
        if not self.interactive:
            return await component.start(['SessionProxy', 'ConsoleUI', 'CoreConfig'])

        await component.start()
        component.pause(
            [
                'TorrentList',
                'EventView',
                'AddTorrents',
                'TorrentDetail',
                'Preferences',
            ]
        )

    @maybe_coroutine
    async def start_console(self):
        self.started_deferred = defer.Deferred()

        if self.initialized:
            await component.stop(['SessionProxy'])
            await component.start(['SessionProxy'])
        else:
            self.initialized = True
            await self.start_components()

    @maybe_coroutine
    async def start(self):
        result = await client.core.get_session_state()
        # Maintain a list of (torrent_id, name) for use in tab completion
        self.torrents = []
        self.events = []

        torrents = await client.core.get_torrents_status({'id': result}, ['name'])
        for torrent_id, status in torrents.items():
            self.torrents.append((torrent_id, status['name']))

        self.started_deferred.callback(True)

        # Register event handlers to keep the torrent list up-to-date
        client.register_event_handler('TorrentAddedEvent', self.on_torrent_added)
        client.register_event_handler('TorrentRemovedEvent', self.on_torrent_removed)

    @defer.inlineCallbacks
    def on_torrent_added(self, event, from_state=False):
        status = yield client.core.get_torrent_status(event, ['name'])
        self.torrents.append((event, status['name']))

    def on_torrent_removed(self, event):
        for index, (tid, name) in enumerate(self.torrents):
            if event == tid:
                del self.torrents[index]

    def match_torrents(self, strings):
        return list(
            {torrent for string in strings for torrent in self.match_torrent(string)}
        )

    def match_torrent(self, string):
        """
        Returns a list of torrent_id matches for the string.  It will search both
        torrent_ids and torrent names, but will only return torrent_ids.

        :param string: str, the string to match on

        :returns: list of matching torrent_ids. Will return an empty list if
            no matches are found.

        """
        deluge.common.decode_bytes(string, self.encoding)

        if string == '*' or string == '':
            return [tid for tid, name in self.torrents]

        match_func = '__eq__'
        if string.startswith('*'):
            string = string[1:]
            match_func = 'endswith'
        if string.endswith('*'):
            match_func = '__contains__' if match_func == 'endswith' else 'startswith'
            string = string[:-1]

        matches = []
        for tid, name in self.torrents:
            deluge.common.decode_bytes(name, self.encoding)
            if getattr(tid, match_func, None)(string) or getattr(
                name, match_func, None
            )(string):
                matches.append(tid)
        return matches

    def get_torrent_name(self, torrent_id):
        for tid, name in self.torrents:
            if torrent_id == tid:
                return name
        return None

    def set_batch_write(self, batch):
        if self.interactive and isinstance(
            self.active_mode, deluge.ui.console.modes.cmdline.CmdLine
        ):
            return self.active_mode.set_batch_write(batch)

    def tab_complete_torrent(self, line):
        if self.interactive and isinstance(
            self.active_mode, deluge.ui.console.modes.cmdline.CmdLine
        ):
            return self.active_mode.tab_complete_torrent(line)

    def tab_complete_path(
        self, line, path_type='file', ext='', sort='name', dirs_first=True
    ):
        if self.interactive and isinstance(
            self.active_mode, deluge.ui.console.modes.cmdline.CmdLine
        ):
            return self.active_mode.tab_complete_path(
                line, path_type=path_type, ext=ext, sort=sort, dirs_first=dirs_first
            )

    def on_client_disconnect(self):
        component.stop()

    def write(self, s):
        if self.interactive:
            if isinstance(self.active_mode, deluge.ui.console.modes.cmdline.CmdLine):
                self.active_mode.write(s)
            else:
                component.get('CmdLine').add_line(s, False)
                self.events.append(s)
        else:
            print(colors.strip_colors(s))

    def write_event(self, s):
        if self.interactive:
            if isinstance(self.active_mode, deluge.ui.console.modes.cmdline.CmdLine):
                self.events.append(s)
                self.active_mode.write(s)
            else:
                component.get('CmdLine').add_line(s, False)
                self.events.append(s)
        else:
            print(colors.strip_colors(s))



#
# Copyright (C) 2010 John Garland <johnnybg+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import inspect
import re
import warnings
from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar

from twisted.internet import defer


def proxy(proxy_func):
    """
    Factory class which returns a decorator that passes
    the decorated function to a proxy function

    :param proxy_func: the proxy function
    :type proxy_func: function
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return proxy_func(func, *args, **kwargs)

        return wrapper

    return decorator


def overrides(*args):
    """
    Decorater function to specify when class methods override
    super class methods.

    When used as
    @overrides
    def funcname

    the argument will be the funcname function.

    When used as
    @overrides(BaseClass)
    def funcname

    the argument will be the BaseClass

    """
    stack = inspect.stack()
    if inspect.isfunction(args[0]):
        return _overrides(stack, args[0])
    else:
        # One or more classes are specified, so return a function that will be
        # called with the real function as argument
        def ret_func(func, **kwargs):
            return _overrides(stack, func, explicit_base_classes=args)

        return ret_func


def _overrides(stack, method, explicit_base_classes=None):
    # stack[0]=overrides, stack[1]=inside class def'n, stack[2]=outside class def'n
    classes = {}
    derived_class_locals = stack[2][0].f_locals

    # Find all super classes
    m = re.search(r'class\s(.+)\((.+)\)\s*\:', stack[2][4][0])
    class_name = m.group(1)
    base_classes = m.group(2)

    # Handle multiple inheritance
    base_classes = [s.strip() for s in base_classes.split(',')]
    check_classes = base_classes

    if not base_classes:
        raise ValueError(
            'overrides decorator: unable to determine base class of class "%s"'
            % class_name
        )

    def get_class(cls_name):
        if '.' not in cls_name:
            return derived_class_locals[cls_name]
        else:
            components = cls_name.split('.')
            # obj is either a module or a class
            obj = derived_class_locals[components[0]]
            for c in components[1:]:
                assert inspect.ismodule(obj) or inspect.isclass(obj)
                obj = getattr(obj, c)
            return obj

    if explicit_base_classes:
        # One or more base classes are explicitly given, check only those classes
        override_classes = re.search(r'\s*@overrides\((.+)\)\s*', stack[1][4][0]).group(
            1
        )
        override_classes = [c.strip() for c in override_classes.split(',')]
        check_classes = override_classes

    for c in base_classes + check_classes:
        classes[c] = get_class(c)

    # Verify that the explicit override class is one of base classes
    if explicit_base_classes:
        from itertools import product

        for bc, cc in product(base_classes, check_classes):
            if issubclass(classes[bc], classes[cc]):
                break
        else:
            raise Exception(
                'Excplicit override class "%s" is not a super class of: %s'
                % (explicit_base_classes, class_name)
            )
        if not all(hasattr(classes[cls], method.__name__) for cls in check_classes):
            for cls in check_classes:
                if not hasattr(classes[cls], method.__name__):
                    raise Exception(
                        'Function override "%s" not found in superclass: %s\n%s'
                        % (
                            method.__name__,
                            cls,
                            f'File: {stack[1][1]}:{stack[1][2]}',
                        )
                    )

    if not any(hasattr(classes[cls], method.__name__) for cls in check_classes):
        raise Exception(
            'Function override "%s" not found in any superclass: %s\n%s'
            % (
                method.__name__,
                check_classes,
                f'File: {stack[1][1]}:{stack[1][2]}',
            )
        )
    return method


def deprecated(func):
    """This is a decorator which can be used to mark function as deprecated.

    It will result in a warning being emitted when the function is used.

    """

    @wraps(func)
    def depr_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # Turn off filter
        warnings.warn(
            f'Call to deprecated function {func.__name__}.',
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter('default', DeprecationWarning)  # Reset filter
        return func(*args, **kwargs)

    return depr_func


class CoroutineDeferred(defer.Deferred):
    """Wraps a coroutine in a Deferred.
    It will dynamically pass through the underlying coroutine without wrapping where apporpriate.
    """

    def __init__(self, coro: Coroutine):
        # Delay this import to make sure a reactor was installed first
        from twisted.internet import reactor

        super().__init__()
        self.coro = coro
        self.awaited = None
        self.activate_deferred = reactor.callLater(0, self.activate)

    def __await__(self):
        if self.awaited in [None, True]:
            self.awaited = True
            return self.coro.__await__()
        # Already in deferred mode
        return super().__await__()

    def activate(self):
        """If the result wasn't awaited before the next context switch, we turn it into a deferred."""
        if self.awaited is None:
            self.awaited = False
            try:
                d = defer.Deferred.fromCoroutine(self.coro)
            except AttributeError:
                # Fallback for Twisted <= 21.2 without fromCoroutine
                d = defer.ensureDeferred(self.coro)
            d.chainDeferred(self)

    def _callback_activate(self):
        """Verify awaited status before calling activate."""
        assert not self.awaited, 'Cannot add callbacks to an already awaited coroutine.'
        self.activate()

    def addCallback(self, *args, **kwargs):  # noqa: N802
        self._callback_activate()
        return super().addCallback(*args, **kwargs)

    def addCallbacks(self, *args, **kwargs):  # noqa: N802
        self._callback_activate()
        return super().addCallbacks(*args, **kwargs)

    def addErrback(self, *args, **kwargs):  # noqa: N802
        self._callback_activate()
        return super().addErrback(*args, **kwargs)

    def addBoth(self, *args, **kwargs):  # noqa: N802
        self._callback_activate()
        return super().addBoth(*args, **kwargs)


_RetT = TypeVar('_RetT')


def maybe_coroutine(
    f: Callable[..., Coroutine[Any, Any, _RetT]],
) -> 'Callable[..., defer.Deferred[_RetT]]':
    """Wraps a coroutine function to make it usable as a normal function that returns a Deferred."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        # Uncomment for quick testing to make sure CoroutineDeferred magic isn't at fault
        # return defer.ensureDeferred(f(*args, **kwargs))
        return CoroutineDeferred(f(*args, **kwargs))

    return wrapper



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import sys

import deluge.common
import deluge.configmanager
import deluge.log
from deluge.argparserbase import ArgParserBase
from deluge.i18n import setup_translation

log = logging.getLogger(__name__)

try:
    from setproctitle import setproctitle
except ImportError:

    def setproctitle(title):
        return


class UI:
    """
    Base class for UI implementations.

    """

    cmd_description = """Override with command description"""

    def __init__(self, name, **kwargs):
        self.__name = name
        self.ui_args = kwargs.pop('ui_args', None)
        setup_translation()
        self.__parser = ArgParserBase(**kwargs)

    def parse_args(self, parser, args=None):
        options = parser.parse_args(args)
        if not hasattr(options, 'remaining'):
            options.remaining = []
        return options

    @property
    def name(self):
        return self.__name

    @property
    def parser(self):
        return self.__parser

    @property
    def options(self):
        return self.__options

    def start(self, parser=None):
        args = sys.argv[1:]
        if parser is None:
            parser = self.parser
        self.__options = self.parse_args(parser, args)

        setproctitle('deluge-%s' % self.__name)

        log.info('Deluge ui %s', deluge.common.get_version())
        log.debug('options: %s', self.__options)
        log.info('Starting %s ui..', self.__name)



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""
This base class is used in plugin's __init__ for the plugin entry points.
"""

import logging

log = logging.getLogger(__name__)


class PluginInitBase:
    _plugin_cls = None

    def __init__(self, plugin_name):
        self.plugin = self._plugin_cls(plugin_name)  # pylint: disable=not-callable

    def enable(self):
        return self.plugin.enable()

    def disable(self):
        return self.plugin.disable()



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component

from . import BaseCommand

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """Enable interactive mode"""

    interactive_only = True

    def handle(self, options):
        console = component.get('ConsoleUI')
        at = console.set_mode('TorrentList')
        at.go_top = True
        at.resume()
